import numpy as np
from numpy.typing import NDArray
from typing import Any, Tuple, Union, List
import sys
from scipy.sparse import spdiags
from scipy import sparse


class Minq:
    """
    The minQ class helper for MCS
    """

    def ldlrk1(self, L: NDArray[np.float64], d: NDArray[np.float64],
               alp: float, u: NDArray[np.float64], eps: float = 2.2204e-16) \
            -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """
        lslrk1 function

        :param L: Indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        :param d: the value of the interpolating polynomial
        :param alp:
        :param u:
        :param eps: parameter value for the golden ratio
        :return:
        """
        p = np.array([])
        if alp == 0:
            return L, d, p

        n = u.shape[0]
        neps = n * eps

        L0 = L
        d0 = d

        for k in [i for i in range(n) if u[i] != 0]:
            delta = d[k] + alp * pow(u[k], 2)
            if alp < 0 and delta <= neps:
                p = np.zeros(n)
                p[k] = 1
                p0Krange = [i for i in range(0, k + 1)]
                p0K = np.asarray([p[i] for i in p0Krange])
                L0K = np.asarray([[L[i, j] for j in p0Krange] for i in p0Krange])
                p0K = np.linalg.solve(L0K, p0K)
                p = np.asarray([p0K[i] if (i in p0Krange) else p[i] for i in range(len(p))])
                L = L0
                d = d0
                return L, d, p

            q = d[k] / delta
            d[k] = delta
            ind = [i for i in range(k + 1, n)]
            LindK = np.asarray([L[i, k] for i in ind])
            uk = u[k]
            c = np.dot(LindK, uk)
            for i in range(len(ind)):
                L[ind[i], k] = LindK[i] * q + (alp * u[k] / delta) * u[ind[i]]

            for i in range(len(ind)):
                u[ind[i]] = u[ind[i]] - c[i]

            alp = alp * q
            if alp == 0:
                break
        return L, d, p


class UtilHelpers:
    """
    A class with util functions for MCS
    """

    def ldldown(self, L: NDArray[np.float64], d: NDArray[np.float64], j: int)  \
            -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        ldldown function

        :param L: indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        :param d: the value of the interpolating polynomial
        :param j: label
        :return: the updated end point and the updated value
        """
        n = d.shape[0]
        if j < n:
            I = [i for i in range(0, j)]
            K = [i for i in range(j + 1, n)]

            LKK = np.asarray([[L[i, j] for j in K] for i in K])
            dK = np.asarray([d[i] for i in K])
            LKj = np.asarray([L[i, j] for i in K])
            LKK, dK, _ = Minq().ldlrk1(LKK, dK, d[j], LKj)
            d[K] = dK
            r1 = L[I, :]
            r2 = sparse.coo_matrix((1, n)).toarray()
            if len(I) == 0:
                r3 = np.concatenate((sparse.coo_matrix((n - j - 1, 1)).toarray(), LKK), axis=1)
                L = np.concatenate((r2, r3), axis=0)
            else:
                LKI = np.asarray([[L[i, j] for j in I] for i in K])
                if len(K) != 0:
                    r3 = np.concatenate((LKI, sparse.coo_matrix((n - j - 1, 1)).toarray(), LKK), axis=1)
                    L = np.concatenate((r1, r2, r3), axis=0)
                else:
                    L = np.concatenate((r1, r2), axis=0)
            L[j, j] = 1
        else:
            L[n - 1, 0: n - 1] = sparse.coo_matrix((1, n - 1)).toarray()
        d[j] = 1
        return L, d

    def ldlup(self, L: NDArray[np.float64], d: NDArray[np.float64], j: int, g: NDArray[np.float64],
              eps: float) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """
        ldlup function
        :param L: Indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        :param d: the value of the interpolating polynomial
        :param j: label
        :param g:
        :param eps: parameter value for the golden ratio
        :return:
        """
        p = np.array([])
        n = d.shape[0]
        I = [i for i in range(0, j)]
        K = [i for i in range(j + 1, n)]

        if j == 0:
            v = np.zeros(0)
            delta = g[j]
            if delta <= n * eps:
                p = np.asarray([1] + np.zeros(n - 1).tolist())
                return L, d, p
            w = np.asarray([g[i] / delta for i in K])
            L[j, I] = v.T
            d[j] = delta
            return L, d, p
        LII = np.asarray([[L[i, j] for j in I] for i in I])
        gI = [g[i] for i in I]
        u = np.linalg.solve(LII, gI)
        dI = [d[i] for i in I]
        v = np.divide(u, dI)
        delta = g[j] - np.dot(u.T, v)
        if delta <= n * eps:
            p = np.asarray(np.linalg.solve(LII.T, v).tolist() + [-1] + np.zeros(n - j - 1).tolist())
            return L, d, p

        if len(K) != 0:
            LKI = np.asarray([[L[i, j] for j in I] for i in K])
            gK = np.asarray([g[i] for i in K])
            w = np.divide(np.subtract(gK, np.dot(LKI, u)), delta)
            LKK = np.asarray([[L[i, j] for j in K] for i in K])
            dK = np.asarray([d[i] for i in K])
            LKK, dK, q = Minq().ldlrk1(LKK, dK, -delta, w)
            d[K] = dK
        else:
            q = np.array([])

        if len(q) == 0:
            r1 = L[I, :]
            r2 = np.asarray(v.T.tolist() + [1] + L[j, K].tolist())
            r2 = r2.reshape((1, len(r2)))
            if len(K) != 0:
                r3 = np.concatenate((LKI, w.reshape(len(w), 1), LKK), axis=1)

                L = np.concatenate((r1, r2, r3), axis=0)
            else:
                L = np.concatenate((r1, r2), axis=0)
            d[j] = delta
        else:
            r1 = L[0: j + 1, :]
            r2 = np.concatenate((LKI, L[K, j].reshape(len(L[K, j]), 1), LKK), axis=1)
            L = np.concatenate((r1, r2), axis=0)
            w = w.reshape((len(w), 1))
            q.reshape((len(q)), 1)
            pi = np.dot(w.T, q)
            piv = np.multiply(pi, v)
            LKIq = np.dot(LKI.T, q)
            pivLKIq = np.subtract(piv.flatten(), LKIq.flatten())
            piSolve = np.linalg.solve(LII.T, pivLKIq)
            p = np.asarray(piSolve.flatten().tolist() + (-1 * pi).flatten().tolist() + q.tolist())
        return L, d, p

    def getalp(self, alpu: float, alpo: float, gTp: float, pTGp: float) -> \
            Tuple[float, bool, bool, int]:
        """
        Gives minimizer alp in [alpu,alpo] for a univariate quadratic q(alp)=alp*gTp+0.5*alp^2*pTGp

        :param alpu:
        :param alpo:
        :param gTp:
        :parm pTGp:
        :return:
        """
        lba = False
        uba = False

        ier = 0
        if alpu == -np.Inf and (pTGp < 0 or (pTGp == 0 and gTp > 0)):
            ier = 1
            lba = True
        if alpo == np.Inf and (pTGp < 0 or (pTGp == 0 and gTp < 0)):
            ier = 1
            uba = True
        if ier:
            alp = np.NAN
            return alp, lba, uba, ier

        if pTGp == 0 and gTp == 0:
            alp = 0
        elif pTGp <= 0:
            if alpu == -np.Inf:
                lba = False
            elif alpo == np.Inf:
                lba = True
            else:
                lba = (2 * gTp + (alpu + alpo) * pTGp > 0)
            uba = not lba
        else:
            alp = -gTp / pTGp
            lba = (alp <= alpu)
            uba = (alp >= alpo)

        if lba:
            alp = alpu
        if uba:
            alp = alpo

        if abs(alp) == np.Inf:
            gTp, pTGp, alpu, alpo, alp, lba, uba, ier
        return alp, lba, uba, ier

    def minqsub(self, nsub: int, free: NDArray[np.bool_], L: NDArray[np.float64],
                dd: NDArray[np.float64], K: NDArray[np.bool_], G: NDArray[np.float64],
                n: int, g: NDArray[np.float64], x: NDArray[np.float64], xo: NDArray[np.float64],
                xu: NDArray[np.float64], convex: int, xx: NDArray[np.float64], fct: float,
                nfree: int, unfix: int, alp: float, alpu: float,
                alpo: float, lba: bool, uba: bool, ier: int, subdone: int, eps: float):
        """
        Minqsub function

        :param nsub:
        :param free:
        :param L: Indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        :param dd:
        :param n:
        :param g:
        :param x:
        :param xo:
        :param xu:
        :param convex:
        :param xx:
        :param fct:
        :param nfree:
        :param unfix:
        :param alp:
        :param alpu:
        :param alpo:
        :param lba:
        :param uba:
        :param ier:
        :param subdone:
        :param eps: parameter value for the golden ratio
        :return:
        """
        nsub = nsub + 1
        freelK = [i for i in range(len(free)) if (free < K)[i] is True]
        for j in freelK:
            L, dd = self.ldldown(L, dd, j)
            K[j] = False

        definite = 1
        freeuK = [i for i in range(len(free)) if (free > K)[i] is True]
        for j in freeuK:
            p = np.zeros(n)
            if n > 1:
                p = np.asarray([G[i, j] if K[i] is True else p[i] for i in range(len(K))])
            p[j] = G[j, j]
            L, dd, p = self.ldlup(L, dd, j, p, eps)
            definite = (len(p) == 0)
            if not definite:
                break
            K[j] = True

        if definite:
            p = np.zeros(n)
            p = np.asarray([g[i] if K[i] is True else p[i] for i in range(len(K))])
            LPsolve = np.linalg.solve(L, p)
            LPsolve = np.divide(LPsolve, dd)
            p = np.multiply(-1, np.linalg.solve(L.T, LPsolve))

        p = (x + p) - x
        ind = [i for i in range(len(p)) if p[i] != 0]
        if len(ind) == 0:
            unfix = 1
            subdone = 0
            return (nsub, free, L, dd, K, G, n, g, x, xo, xu, convex, xx, fct,
                    nfree, alp, alpu, alpo, lba, uba, ier, unfix, subdone)
        pp = np.asarray([p[i] for i in ind])
        oo = np.subtract([xo[i] for i in ind], [x[i] for i in ind]) / pp
        uu = np.subtract([xu[i] for i in ind], [x[i] for i in ind]) / pp
        alpu = max([oo[i] for i in range(len(ind)) if pp[i] < 0] + [uu[i] for i in range(
            len(ind)) if pp[i] > 0] + [-np.inf])
        alpo = min([oo[i] for i in range(len(ind)) if pp[i] > 0] + [uu[i] for i in range(
            len(ind)) if pp[i] < 0] + [np.inf])
        if alpo <= 0 or alpu >= 0:
            sys.exit('programming error: no alp')

        gTp = np.dot(g.T, p)
        agTp = np.dot(np.abs(g).T, np.abs(p))
        if abs(gTp) < 100 * eps * agTp:
            gTp = 0
        pTGp = np.dot(p.T, np.dot(G, p))
        if convex:
            pTGp = max(0, pTGp)
        if not definite and pTGp > 0:
            pTGp = 0

        alp, lba, uba, ier = self.getalp(alpu, alpo, gTp, pTGp)
        if ier:
            x = np.zeros(n)
            if lba:
                x = -p
            else:
                x = p
            return (nsub, free, L, dd, K, G, n, g, x, xo, xu, convex, xx,
                    fct, nfree, alp, alpu, alpo, lba, uba, ier, unfix, subdone)

        unfix = not (lba or uba)
        for k in range(0, len(ind)):
            ik = ind[k]
            if alp == uu[k]:
                xx[ik] = xu[ik]
                free[ik] = 0
            elif alp == oo[k]:
                xx[ik] = xo[ik]
                free[ik] = 0
            else:
                xx[ik] = xx[ik] + alp * p[ik]
            if abs(xx[ik]) == np.Inf:
                ik, alp, p[ik]
                sys.exit('infinite xx in minq')

        nfree = sum(free)
        subdone = 1
        return (nsub, free, L, dd, K, G, n, g, x, xo, xu, convex, xx,
                fct, nfree, alp, alpu, alpo, lba, uba, ier, unfix, subdone)


class LSUtils(UtilHelpers):
    """
    Class with utility functions for MCS
    """

    def lsguard(self, alp: float, alist: List[Union[float, int]], amax: float, amin: float, small: Union[float, int]) \
            -> float:
        """
        Local search guard function

        :param alp:
        :param alist: list of known steps
        :param amax: maximum step
        :param amin: minimum step
        :param small:
        :return:
        """
        asort = alist
        asort.sort()
        s = len(asort)

        al = asort[0] - (asort[s - 1] - asort[0]) / small
        au = asort[s - 1] + (asort[s - 1] - asort[0]) / small
        alp = max(al, min(alp, au))
        alp = max(amin, min(alp, amax))

        if abs(alp - asort[0]) < small * (asort[1] - asort[0]):
            alp = (2 * asort[0] + asort[1]) / 3

        if abs(alp - asort[s - 1]) < small * (asort[s - 1] - asort[s - 1 - 1]):
            alp = (2 * asort[s - 1] + asort[s - 1 - 1]) / 3

        return alp

    def lssplit(self, i: int, alist: List[Union[float, int]], flist: List[Union[float, int]], short: float) \
            -> Tuple[float, float]:
        """
        Local search split function

        :param i: label index
        :param alist: list of known steps
        :param flist: function values of known steps
        :param short:
        :return:
        """
        if flist[i] < flist[i + 1]:
            fac = short
        elif flist[i] > flist[i + 1]:
            fac = 1 - short
        else:
            fac = 0.5

        alp = alist[i] + fac * (alist[i + 1] - alist[i])
        return alp, fac

    def minq(self, gam: float, c: NDArray[np.float64], G: NDArray[np.float64], xu: NDArray[np.float64],
             xo: NDArray[np.float64], prt: int, eps: float, ier: int = 0, convex: int = 0) \
            -> Tuple[NDArray[np.float64], float, int]:
        """
        Minq Function
        Minimizes an affine quadratic form subject to simple bounds.
        Using coordinate searches and reduced subspace minimizations, using LDL^T factorization updates
        fct  =  gam + c^T x + 0.5 x^T G x s.t. x in [xu,xo] (xu <= xo is assumed),
        where G is a symmetric n x n matrix, not necessarily definite
        (if G is indefinite, only a local minimum is found)
        if G is sparse, it is assumed that the ordering is such that
        a sparse modified Cholesky factorization is feasible

        :param prt:	print command
        :param xx: initial guess (optional)
        :param x: minimizer (but unbounded direction if ier = 1)
        :param fct:	optimal function value
        :param ier:	0 (local minimizer found)
        """

        n = G.shape[0]

        if G.shape[1] != n:
            ier = -1
            print('minq: Hessian has wrong dimension')
            x = np.NAN + np.zeros(n)
            fct = np.NAN
            nsub = -1
            return x, fct, ier

        if c.shape[0] != n:
            ier = -1
            print('minq: linear term has wrong dimension')
        if xu.shape[0] != n:
            ier = -1
            print('minq: lower bound has wrong dimension')

        if xo.shape[0] != n:
            ier = -1
            print('minq: lower bound has wrong dimension')

        if 'xx' in locals():
            xx: NDArray[Any] = locals()["xx"]
            if xx.shape[0] != n:
                ier = -1
                print('minq: lower bound has wrong dimension')
        else:
            xx = np.zeros(n)

        if ier == -1:
            x = np.NAN + np.zeros(n)
            fct = np.NAN
            nsub = -1
            return x, fct, ier

        maxit = 3 * n

        nitrefmax = 3
        xx = np.asarray([max(xu[i], min(xx[i], xo[i])) for i in range(len(xx))])

        hpeps = 100 * eps
        G = G + spdiags(hpeps * np.diag(G), 0, n, n).toarray()

        K = np.zeros(n, dtype=bool)
        L = np.eye(n)
        dd = np.ones(n)

        free = np.zeros(n, dtype=bool)
        nfree = 0
        nfree_old = -1

        fct = np.Inf
        nsub = 0
        unfix = 1
        nitref = 0
        improvement = 1

        while 1:
            if np.linalg.norm(xx, np.inf) == np.inf:
                sys.exit('infinite xx in minq.m')

            g = np.dot(G, xx) + c
            fctnew = gam + np.dot(0.5 * xx.T, (c + g))
            if not improvement:
                ier = 0
                break
            elif nitref > nitrefmax:

                ier = 0
                break
            elif nitref > 0 and nfree_old == nfree and fctnew >= fct:
                ier = 0
                break
            elif nitref == 0:
                x = xx
                fct = min(fct, fctnew)
            else:
                x = xx
                fct = fctnew
            if nitref == 0 and nsub >= maxit:
                ier = 99
                break
            count = 0
            k = -1
            while 1:
                while count <= n:
                    count = count + 1
                    if k == n - 1:
                        k = -1
                    k = k + 1
                    if free[k] or unfix:
                        break
                if count > n:
                    break
                q = G[:, k]
                alpu = xu[k] - x[k]
                alpo = xo[k] - x[k]

                alp, lba, uba, ier = self.getalp(alpu, alpo, g[k], q[k])

                if ier:
                    x = np.zeros(n)
                    if lba:
                        x[k] = -1
                    else:
                        x[k] = 1

                    return x, fct, ier

                xnew = x[k] + alp
                if prt and nitref > 0:
                    xnew, alp

                if lba or xnew <= xu[k]:
                    if alpu != 0:
                        x[k] = xu[k]
                        g = g + alpu * q
                        count = 0
                    free[k] = 0
                elif uba or xnew >= xo[k]:
                    if alpo != 0:
                        x[k] = xo[k]
                        g = g + alpo * q
                        count = 0
                    free[k] = 0
                else:
                    if alp != 0.0:
                        if prt > 1 and not free[k]:
                            unfixstep = [x[k], alp]
                        x[k] = xnew
                        g = g + alp * q
                        free[k] = 1

            nfree = sum(free)
            if (unfix and nfree_old == nfree):
                g = np.dot(G, x) + c
                nitref = nitref + 1
            else:
                nitref = 0
            nfree_old = nfree
            gain_cs = fct - gam - np.dot(0.5 * x.T, (c + g))
            improvement = (gain_cs > 0 or not unfix)
            xx = x
            if not improvement or nitref > nitrefmax:

                nothing_to_do = 'done!'
            elif nitref > nitrefmax:
                nothing_to_do = 'done!'
            elif nfree == 0:
                unfix = 1
            else:
                subdone = 0
                (nsub, free, L, dd, K, G, n, g, x, xo, xu, convex, xx, fct,
                 nfree, alp, alpu, alpo, lba, uba, ier, unfix, subdone) = self.minqsub(nsub, free, L, dd, K, G, n, g, x,
                                                                                       xo, xu, convex, xx, fct, nfree,
                                                                                       unfix, alp, alpu, alpo, lba, uba,
                                                                                       ier, subdone, eps)
                if not subdone:
                    return x, fct, ier
                if ier:
                    return x, fct, ier
        return x, fct, ier

    def quartic(self, a: NDArray[np.float64], x: float):
        """
        quartic function

        :param a: vector of function values, normalized by its maximum value
        :param x: positional argument generated from alp
        :return: scalar value adjused for quart bool in main code (mcs_agent)
        """
        return (((a[0] * x + a[1]) * x + a[2]) * x + a[3]) * x + a[4]



"""
MIT License

Copyright (c) 2019 CleanRL developers https://github.com/vwxyzjn/cleanrl
"""
from typing import List, Optional, Tuple, Any, Callable
import time
import torch
import torch.optim as optim
import gymnasium as gym
import os
import random
import torch.nn.functional as F
import numpy as np
from stable_baselines3.common.buffers import ReplayBuffer
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from gymnasium.wrappers.common import RecordEpisodeStatistics
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.models.q_network import QNetwork
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.dao.training.dqn_policy import DQNPolicy
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_agents.agents.base.base_agent import BaseAgent
import csle_agents.constants.constants as agents_constants


class DQNCleanAgent(BaseAgent):
    """
    A DQN agent using the implementation from OpenAI baselines
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 emulation_env_config: Optional[EmulationEnvConfig], experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True) -> None:
        """
        Initializes the agent

        :param simulation_env_config: the simulation environment configuration
        :param emulation_env_config: the emulation environment configuration
        :param experiment_config: the experiment configuration
        :param training_job: the training job configuration
        :param save_to_metastore: boolean flag indicating whether job information should be saved to the metastore
        """
        super(DQNCleanAgent, self).__init__(simulation_env_config=simulation_env_config,
                                            emulation_env_config=emulation_env_config,
                                            experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.DQN_CLEAN
        self.training_job = training_job
        self.num_hl = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS].value
        self.num_hl_neur = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER].value
        self.config = self.simulation_env_config.simulation_env_input_config
        self.save_to_metastore = save_to_metastore
        self.exp_name: str = self.simulation_env_config.name
        self.env_id = self.simulation_env_config.gym_env_name
        self.torch_deterministic = True
        self.cuda: bool = True
        self.learning_rate = self.experiment_config.hparams[agents_constants.COMMON.LEARNING_RATE].value
        self.start_e = 1
        self.end_e = 0.05
        self.num_envs = self.experiment_config.hparams[agents_constants.COMMON.NUM_PARALLEL_ENVS].value
        self.total_timesteps = self.experiment_config.hparams[agents_constants.COMMON.NUM_TRAINING_TIMESTEPS].value
        self.batch_size = self.experiment_config.hparams[agents_constants.COMMON.BATCH_SIZE].value
        self.num_iterations = self.total_timesteps // self.batch_size
        self.gamma = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        self.tau = self.experiment_config.hparams[agents_constants.DQN_CLEAN.TAU].value
        self.exploration_fraction = self.experiment_config.hparams[agents_constants.DQN_CLEAN.EXP_FRAC].value
        self.learning_starts = self.experiment_config.hparams[agents_constants.DQN_CLEAN.LEARNING_STARTS].value
        self.train_frequency = self.experiment_config.hparams[agents_constants.DQN_CLEAN.TRAIN_FREQ].value
        self.target_network_frequency = self.experiment_config.hparams[agents_constants.DQN_CLEAN.T_N_FREQ].value
        self.buffer_size = self.experiment_config.hparams[agents_constants.DQN_CLEAN.BUFFER_SIZE].value
        self.save_model = self.experiment_config.hparams[agents_constants.DQN_CLEAN.SAVE_MODEL].value
        self.device = self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value
        self.orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=self.config)

    def linear_schedule(self, duration: int, t: int) -> float:
        """
        Linear exploration rate decay sechdule

        :param duration: the duration of training
        :param t: the current time
        :return: the new exploration rate
        """
        slope = (self.end_e - self.start_e) / duration
        return max(slope * t + self.start_e, self.end_e)

    def make_env(self) -> Callable[[], RecordEpisodeStatistics[Any, Any]]:
        """
        Helper function for creating an environment in training of the agent

        :return: environment creating function
        """

        def thunk() -> RecordEpisodeStatistics[Any, Any]:
            """
            The environment creating function
            """
            orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=self.config)
            env = RecordEpisodeStatistics(orig_env)
            return env

        return thunk

    def train(self) -> ExperimentExecution:
        """
        Implements the training logic of the DQN algorithm

        :return: the experiment result
        """
        pid = os.getpid()

        # Setup experiment metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RANDOM_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNTIME)
        descr = f"Training of policies with Clean-DQN using " \
                f"simulation:{self.simulation_env_config.name}"

        # Setup training job
        if self.training_job is None:
            emulation_name = ""
            if self.emulation_env_config is not None:
                emulation_name = self.emulation_env_config.name
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=emulation_name, simulation_traces=[],
                num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            training_job_id = -1
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
            self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Setup experiment execution
        ts = time.time()
        simulation_name = self.simulation_env_config.name
        emulation_name = ""
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        self.exp_execution = ExperimentExecution(
            result=exp_result, config=self.experiment_config, timestamp=ts,
            emulation_name=emulation_name, simulation_name=simulation_name,
            descr=descr, log_file_path=self.training_job.log_file_path)
        exp_execution_id = -1
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
        self.exp_execution.id = exp_execution_id

        # Training runs, one per seed
        for seed in self.experiment_config.random_seeds:
            assert self.num_envs == 1, "vectorized envs are not supported at the moment"
            # Train
            exp_result, env, model = self.run_dqn(exp_result=exp_result, seed=seed)

            # Save policy
            ts = time.time()
            save_path = f"{self.experiment_config.output_dir}/DQN_policy_seed_{seed}_{ts}.zip"
            model.save(save_path)
            action_space = \
                self.simulation_env_config.joint_action_space_config.action_spaces[
                    self.experiment_config.player_idx].actions
            policy = DQNPolicy(model=model, simulation_name=self.simulation_env_config.name,
                               save_path=save_path, player_type=self.experiment_config.player_type,
                               states=self.simulation_env_config.state_space_config.states,
                               actions=action_space,
                               experiment_config=self.experiment_config,
                               avg_R=exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN][-1])

            if self.save_to_metastore:
                MetastoreFacade.save_dqn_policy(dqn_policy=policy)
                os.chmod(save_path, 0o777)

            # Save trace
            traces = env.get_traces()
            if len(traces) > 0 and self.save_to_metastore:
                MetastoreFacade.save_simulation_trace(traces[-1])
            env.reset_traces()

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                seed_values = []
                for seed_idx in range(len(self.experiment_config.random_seeds)):
                    seed_values.append(value_vectors[seed_idx][i])
                avg_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0])
                std_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1])
            exp_result.avg_metrics[metric] = avg_metrics
            exp_result.std_metrics[metric] = std_metrics
        return self.exp_execution

    def run_dqn(self, exp_result: ExperimentResult, seed: int) -> Tuple[ExperimentResult, BaseEnv, QNetwork]:
        """
        Runs DQN with given seed

        :param exp_result: the object to save the experiment results
        :param seed: the random seed
        :retur: the updated experiment results, the environment and the trained model
        """
        Logger.__call__().get_logger().info(f"[CleanDQN] Start training; seed: {seed}")
        envs = gym.vector.SyncVectorEnv([self.make_env() for _ in range(self.num_envs)])
        assert isinstance(envs.single_action_space, gym.spaces.Discrete), "only discrete action space is supported"

        # Setup training metrics
        exp_result.all_metrics[seed] = {}
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME] = []
        ExperimentUtil.set_seed(seed)
        cuda = False

        # Create neural network
        device = torch.device(agents_constants.DQN_CLEAN.CUDA if torch.cuda.is_available() and cuda else
                              self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value)
        input_dim = np.array(envs.single_observation_space.shape).prod()
        q_network = QNetwork(input_dim=input_dim, num_hidden_layers=self.num_hl,
                             hidden_layer_dim=self.num_hl_neur, agent_type=self.experiment_config.agent_type).to(device)
        optimizer = optim.Adam(q_network.parameters(), lr=self.learning_rate)
        target_network = QNetwork(input_dim=input_dim, num_hidden_layers=self.num_hl,
                                  hidden_layer_dim=self.num_hl_neur,
                                  agent_type=self.experiment_config.agent_type).to(device)

        # Seeding
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = self.torch_deterministic

        target_network.load_state_dict(q_network.state_dict())
        rb = ReplayBuffer(
            self.buffer_size,
            envs.single_observation_space,
            envs.single_action_space,
            device,
            handle_timeout_termination=False,
        )
        start_time = time.time()

        obs: List[float] = envs.reset(seed=seed)[0]
        R: List[Any] = []
        T = []
        i = 0
        for global_step in range(self.total_timesteps):
            i += 1
            # ALGO LOGIC: put action logic here
            epsilon = self.linear_schedule(self.exploration_fraction * self.total_timesteps, global_step)
            if random.random() < epsilon:
                actions = np.array([envs.single_action_space.sample() for _ in range(envs.num_envs)])
            else:

                q_values = q_network(torch.Tensor(obs).to(device))
                actions = torch.argmax(q_values, dim=1).cpu().numpy()
                e_name = "csle-stopping-game-pomdp-defender-v1"
                if self.simulation_env_config.simulation_env_input_config.env_name == e_name:
                    actions[0] = random.randrange(0, 2)

            # Take a step in the environment
            next_obs, rewards, terminations, truncations, infos = envs.step(actions) # type: ignore

            # Record rewards for logging purposes
            if "final_info" in infos:
                R_sum = 0
                T_sum = 0
                for info in infos["final_info"]:
                    R_sum += info["R"]
                    T_sum += info["T"]
                R.append(R_sum)
                T.append(T_sum)

            # Save data to replay buffer
            real_next_obs = next_obs.copy()
            for idx, trunc in enumerate(truncations):
                if trunc and "final_observation" in infos:
                    real_next_obs[idx] = infos["final_observation"][idx]
            rb.add(obs, real_next_obs, actions, rewards, terminations, infos)  # type: ignore

            # Move to the next observation
            obs = next_obs

            # Optimizing the neural network based on data from the replay buffer
            if global_step > self.learning_starts:
                if global_step % self.train_frequency == 0:
                    data = rb.sample(self.batch_size)
                    with torch.no_grad():
                        target_max, _ = target_network(data.next_observations.to(dtype=torch.float32)).max(dim=1)
                        td_target = data.rewards.flatten() + self.gamma * target_max * (1 - data.dones.flatten())
                    old_val = q_network(data.observations.to(dtype=torch.float32)).gather(1, data.actions).squeeze()
                    loss = F.mse_loss(td_target, old_val)

                    # optimize the model
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()

                # update target network
                if global_step % self.target_network_frequency == 0:
                    for target_network_param, q_network_param in zip(target_network.parameters(),
                                                                     q_network.parameters()):
                        target_network_param.data.copy_(
                            self.tau * q_network_param.data + (1.0 - self.tau) * target_network_param.data
                        )
            # Logging
            if global_step > 10 and global_step % self.experiment_config.log_every == 0:
                time_elapsed_minutes = round((time.time() - start_time) / 60, 3)
                exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
                avg_R = round(float(np.mean(R)), 3)
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN].append(round(avg_R, 3))
                avg_T = round(float(np.mean(T)), 3)
                exp_result.all_metrics[seed][
                    agents_constants.COMMON.AVERAGE_TIME_HORIZON].append(round(avg_T, 3))
                running_avg_J = ExperimentUtil.running_average(
                    exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN],
                    self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value
                )
                running_avg_T = ExperimentUtil.running_average(
                    exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON],
                    self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
                exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN].append(
                    round(running_avg_J, 3))
                Logger.__call__().get_logger().info(
                    f"[CleanDQN] Iteration: {global_step}/{self.total_timesteps}, "
                    f"avg R: {avg_R}, "
                    f"R_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}:"
                    f"{running_avg_J}, Avg T:{round(avg_T, 3)}, "
                    f"Running_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}_T: "
                    f"{round(running_avg_T, 3)}, "
                    f"runtime: {time_elapsed_minutes} min")
        envs.close()

        base_env: BaseEnv = envs.envs[0].env.env.env  # type: ignore
        return exp_result, base_env, q_network

    def hparam_names(self) -> List[str]:
        """
        Gets the hyperparameters

        :return: a list with the hyperparameter names
        """
        return [constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER,
                constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS,
                agents_constants.COMMON.LEARNING_RATE, agents_constants.COMMON.BATCH_SIZE,
                agents_constants.COMMON.GAMMA,
                agents_constants.COMMON.NUM_TRAINING_TIMESTEPS, agents_constants.COMMON.EVAL_EVERY,
                agents_constants.COMMON.EVAL_BATCH_SIZE,
                constants.NEURAL_NETWORKS.DEVICE,
                agents_constants.COMMON.SAVE_EVERY, agents_constants.DQN.EXPLORATION_INITIAL_EPS,
                agents_constants.DQN.EXPLORATION_FINAL_EPS, agents_constants.DQN.EXPLORATION_FRACTION,
                agents_constants.DQN.MLP_POLICY, agents_constants.DQN.MAX_GRAD_NORM,
                agents_constants.DQN.GRADIENT_STEPS, agents_constants.DQN.N_EPISODES_ROLLOUT,
                agents_constants.DQN.TARGET_UPDATE_INTERVAL, agents_constants.DQN.LEARNING_STARTS,
                agents_constants.DQN.BUFFER_SIZE, agents_constants.DQN.DQN_BATCH_SIZE]



from typing import List, Union, Any, Dict
from csle_agents.agents.pomcp.node import Node
from csle_agents.agents.pomcp.action_node import ActionNode
from csle_agents.agents.pomcp.belief_node import BeliefNode


class BeliefTree:
    """
    The belief tree of POMCP. Each node in the tree corresponds to a history of the POMDP, where a history is a sequence
    of actions and observations.
    """

    def __init__(self, root_particles: List[int], default_node_value: float, root_observation: int,
                 initial_visit_count: int = 0) -> None:
        """
        Initializes the tree with a belief node with a set of particles

        :param root_particles: the particles to add to the root belief node
        :param default_node_value: the default value of nodes in the tree
        :param root_observation: the root observation
        :param initial_visit_count: the initial visit count
        :return: None
        """
        self.tree_size = 0
        self.root_observation = root_observation
        self.nodes: Dict[int, Union[Node, None]] = {}
        self.default_node_value = default_node_value
        self.initial_visit_count = initial_visit_count
        node: Node = self.add(history=[root_observation], particle=root_particles, parent=None,
                              value=default_node_value)
        if isinstance(node, BeliefNode):
            self.root: BeliefNode = node
        else:
            raise ValueError("Invalid root node")

    def add(self, history: List[int], parent: Union[Node, ActionNode, BeliefNode, None],
            action: Union[int, None] = None, observation: Union[int, None] = None, particle: Union[Any, None] = None,
            value: float = 0, initial_visit_count: int = 0) -> Node:
        """
        Creates and adds a new belief node or action node to the belief search tree

        :param h: history sequence
        :param parent: either ActionNode or BeliefNode
        :param action: action
        :param observation: observation
        :param particle: new node's particle set
        :param cost: action cost of an action node
        :param value: the value of the node
        :param initial_visit_count: the initial visit count
        :return: The newly added node
        """
        # Create the node
        if action is not None:
            new_node: Node = ActionNode(self.tree_size, history, parent=parent, action=action, value=value,
                                        visit_count=initial_visit_count)
        else:
            if observation is None:
                observation = 0
            new_node = BeliefNode(self.tree_size, history, parent=parent, observation=observation, value=value,
                                  visit_count=initial_visit_count)

        if particle is not None and isinstance(new_node, BeliefNode):
            new_node.add_particle(particle)

        # add the node to belief tree
        self.nodes[new_node.id] = new_node
        self.tree_size += 1

        # register node as parent's child
        if parent is not None:
            parent.add_child(node=new_node)
        return new_node

    def find_or_create(self, history: List[int], parent: Union[None, BeliefNode, ActionNode], observation: int,
                       initial_value: float, initial_visit_count: int) -> Node:
        """
        Search for the node that corresponds to given history, otherwise create one using given params

        :param history: the current history
        :param parent: the parent of the node
        :param observation: the latest observation
        :param initial_value: the initial value of a created node
        :param initial_visit_count: the initial visit count of a created node
        :return: the new node
        """
        # Start the search from the root node
        root_node = self.root
        current_node: Union[None, Node] = root_node

        # Start from the root node and then traverse down to the depth of the given history to see if the node
        # of this history  exists or not, otherwise add it
        history_length, root_history_length = len(history), len(self.root.history)
        for step in range(root_history_length, history_length):
            if current_node is not None:
                current_node = current_node.get_child(history[step])

            # Node of this history does not exists so we add it
            if current_node is None:
                return self.add(history=history, parent=parent, observation=observation, value=initial_value,
                                initial_visit_count=initial_visit_count)
        if current_node is None:
            raise ValueError("Could not create a new node")
        return current_node

    def prune(self, node, exclude=None):
        """
        Removes the entire subtree subscribed to 'node' with exceptions.
        :param node: root of the subtree to be removed
        :param exclude: exception component
        :return:
        """
        for child in node.children:
            if exclude and exclude.id != child.id:
                self.prune(child, exclude)

        self.nodes[node.id] = None
        del self.nodes[node.id]



import numpy as np
import copy
from typing import Union, List, Tuple, Any
import math
from numpy.typing import NDArray
import sys


class UtilHelpers:
    """
    Class with utility functions for MCS
    """

    def polint(self, x: Union[List[float], NDArray[np.float64]],
               f: Union[List[float], NDArray[np.float64], NDArray[np.float32]]) -> NDArray[np.float64]:
        """
        Quadratic polynomial interpolation

        :param x: pairwise distinct support points
        :param f: corresponding function values
        :return d: the value of the interpolating polynomial
        """
        d = np.zeros(3)
        d[0] = f[0]
        d[1] = (f[1] - f[0]) / (x[1] - x[0])
        f12 = (f[2] - f[1]) / (x[2] - x[1])
        d[2] = (f12 - d[1]) / (x[2] - x[0])
        return d

    def subint(self, x1: float, x2: float) -> Tuple[float, float]:
        """
        Computes [min(x,y),max(x,y)] that are neither too close nor too far away from x

        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        """
        f: int = 1000
        if f * abs(x1) < 1:
            if abs(x2) > f:
                x2 = np.sign(x2)
        else:
            if abs(x2) > f:
                x2 = 10 * np.sign(x2) * abs(x1)
        x1 = x1 + (x2 - x1) / 10
        return x1, x2

    def quadpol(self, x: float, d: NDArray[np.float64], x0: Union[List[float], NDArray[np.float64]]):
        """
        Evaluates the quadratic polynomial

        :param x: starting point
        :param d: the value of the interpolating polynomial
        :param x0: initial position
        """
        return d[0] + d[1] * (x - x0[0]) + d[2] * (x - x0[0]) * (x - x0[1])

    def quadmin(self, a: float, b: float, d: NDArray[np.float64], x0: Union[List[float], NDArray[np.float64]]) -> float:
        """
        The quadmin method

        :param a:
        :param b:
        :param d:
        :param x0:
        :return:
        """
        if d[2] == 0:
            if d[1] > 0:
                x = a
            else:
                x = b
        elif d[2] > 0:
            x1 = 0.5 * (x0[0] + x0[1]) - 0.5 * d[1] / d[2]
            if a <= x1 and x1 <= b:
                x = x1
            elif self.quadpol(a, d, x0) < self.quadpol(b, d, x0):
                x = a
            else:
                x = b
        else:
            if self.quadpol(a, d, x0) < self.quadpol(b, d, x0):
                x = a
            else:
                x = b
        return x

    def split1(self, x1: float, x2: float, f1: float, f2: float) -> float:
        """
        The split1 method

        :param x1:
        :param x2:
        :param f1:
        :param f2:
        :return:
        """
        if f1 <= f2:
            return x1 + 0.5 * (-1 + math.sqrt(5)) * (x2 - x1)
        else:
            return x1 + 0.5 * (3 - math.sqrt(5)) * (x2 - x1)

    def split2(self, x: float, y: float) -> float:
        """
        The split2 method. Determines a value x1 for splitting the interval [min(x,y),max(x,y)]
        is modeled on the function subint with safeguards for infinite y

        :param x:
        :param y:
        :return:
        """
        x2 = y
        if x == 0 and abs(y) > 1000:
            x2 = np.sign(y)
        elif x != 0 and abs(y) > 100 * abs(x):
            x2 = 10 * np.sign(y) * abs(x)
        x1 = x + 2 * (x2 - x) / 3
        return x1

    def vert1(self, j: int, z: NDArray[np.float64], f: NDArray[np.float64], x1: float,
              x2: float, f1: float, f2: float) \
            -> Tuple[float, float, float, float, float]:
        """
        The vert1 method

        :param j: label
        :param z:
        :param f: function value
        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        :param f1: corresponding function value
        :param f2: corresponding function value
        :return:
        """
        if j == 0:
            j1 = 1
        else:
            j1 = 0
        x = z[j1]
        if x1 == np.Inf:
            x1 = z[j]
            f1 = f1 + f[j]
        elif x2 == np.Inf and x1 != z[j]:
            x2 = z[j]
            f2 = f2 + f[j]

        return x, x1, x2, f1, f2

    def vert2(self, j: int, x: float, z: NDArray[np.float64],
              f: NDArray[np.float64], x1: float, x2: float,
              f1: float, f2: float) -> Tuple[float, float, float, float]:
        """
        The vert2 function

        :param j: label
        :param x:
        :param z:
        :param f: function values
        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        :param f1: corresponding function value
        :param f2: corresponding function value
        """
        if j == 0:
            j1 = 1
        else:
            j1 = 0

        if x1 == np.Inf:
            x1 = z[j]
            f1 = f1 + f[j]
            if x != z[j1]:
                x2 = z[j1]
                f2 = f2 + f[j1]
        elif x2 == np.Inf and x1 != z[j]:
            x2 = z[j]
            f2 = f2 + f[j]
        elif x2 == np.Inf:
            x2 = z[j1]
            f2 = f2 + f[j1]

        return x1, x2, f1, f2

    def vert3(self, j: int, x0, f0, L: int, x1: float, x2: float, f1: float, f2: float) \
            -> Tuple[float, float, float, float]:
        """
        Vert3 function

        :param j: label
        :param x0: initial position
        :param f0: inital function value
        :param L:
        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        :param f1: corresponding function value
        :param f2: corresponding function value
        """
        if j == 0:
            k1 = 1
            k2 = 2
        elif j == L:
            k1 = L - 2
            k2 = L - 1
        else:
            k1 = j - 1
            k2 = j + 1
        x1 = x0[k1]
        x2 = x0[k2]
        f1 = f1 + f0[k1]
        f2 = f2 + f0[k2]
        return x1, x2, f1, f2

    def updtf(self, n: int, i: int, x1: NDArray[np.float64], x2: NDArray[np.float64], f1: NDArray[np.float64],
              f2: NDArray[np.float64], fold: Union[float, NDArray[np.float64]], f: NDArray[np.float64]) \
            -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """
        updtf function

        :param n:
        :param i:
        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        :param f1: corresponding function value
        :param f2: corresponding function value
        :param fold: former function value
        :param f: function values
        :return:
        """
        for i1 in range(n):
            if i1 != i:
                if x1[i1] == np.Inf:
                    f1[i1] = f1[i1] + fold - f
                if x2[i1] == np.Inf:
                    f2[i1] = f2[i1] + fold - f
        fold = f
        return f1, f2, fold


class MCSUtils(UtilHelpers):
    """
    Class with utiltiy functions for MCS
    """

    def check_box_bound(self, u: List[int], v: List[int]):
        """
        Function that checks the bounds of the box
        :param u: lower bound
        :param v: upper bound
        :return: boolean indicating the bound
        """
        if v < u:
            print('incompatible box bounds')
            return True
        elif (u == v):
            print('degenerate box bound')
            return True
        else:
            return False

    def strtsw(self, smax: int, level: List[int], f: List[float], nboxes: int, record: NDArray[Any]) \
            -> Tuple[int, NDArray[np.int32]]:
        """
        Function that does the strtsw

        :param smax:
        :param level:
        :param f:
        :param nboxes: counter for boxes not in the 'shopping bas
        :param record:
        """
        record = np.zeros(smax).astype(int)
        s = smax
        for j in range(nboxes + 1):
            if level[j] > 0:
                if level[j] < s:
                    s = level[j]
                if not record[level[j]]:
                    record[level[j]] = j

                elif f[j] < f[record[level[j]]]:
                    record[level[j]] = j
        return s, record

    def exgain(self, n: int, n0: NDArray[np.float64], l: NDArray[np.int32], L: NDArray[np.int32],
               x: NDArray[np.float64], y: NDArray[np.float32], x1: NDArray[np.float32],
               x2: NDArray[np.float32], fx: float, f0: NDArray[np.float32],
               f1: NDArray[np.float32], f2: NDArray[np.float32]) -> Tuple[NDArray[np.float64], int, float]:
        """
        Determines the splitting index, the splitting value and the expected
        gain vector e for (potentially) splitting a box by expected gain

        :param n: dimension of the problem
        :param n0: the ith coordinate has been split n0(i) times in the history of the box
        :param l: Pointer to the initial point of the initialization list
        :param L: lengths of the initialization list
        :param x: base vertex of the box
        :param y: opposite vertex of the box
        :param x1: corresponding parameter/coordinate value
        :param x2: corresponding parameter/coordinate value
        :param f1: corresponding function value
        :param f2: corresponding function value
        :param fx: function value at the base vertex
        :param f0:  function values appertaining to the init. list
        :return e: maximal expected gain in function value by changing coordinate i
        :return isplit: splitting index
        :return splval: Inf  if n0(isplit) = 0, splitting value  otherwise
        """
        e = np.zeros(n)
        emin = np.Inf
        for i in range(n):
            if n0[i] == 0:
                e[i] = min(f0[0: L[i] + 1, i]) - f0[l[i], i]
                if e[i] < emin:
                    emin = e[i]
                    isplit = i
                    splval = np.Inf
            else:
                z1 = [x[i], x1[i], x2[i]]
                z2 = [0, f1[i] - fx, f2[i] - fx]
                d = self.polint(z1, z2)
                eta1, eta2 = self.subint(x[i], y[i])
                xi1 = min(eta1, eta2)
                xi2 = max(eta1, eta2)
                z = self.quadmin(xi1, xi2, d, z1)
                e[i] = self.quadpol(z, d, z1)
                if e[i] < emin:
                    emin = e[i]
                    isplit = i
                    splval = z
        return e, isplit, splval

    def updtrec(self, j: int, s: int, f: List[float], record: List[int]) -> List[int]:
        """
        Updates the pointer record(s) to the best non-split box at level s
        :param j: label of a box
        :param s: its level
        :param f: vector containing the base vertex function values of the already defined boxes.
        :param record: record list
        """
        if len(record) < s:
            record[s] = j
        elif record[s] == 0:
            record[s] = j
        elif f[j] < f[record[s]]:
            record[s] = j

        return record

    def chkloc(self, nloc: int, xloc: List[float], x: float) -> int:
        """
        Checking the location

        :param nloc:
        :param xloc:
        :param x:
        :return: the location
        """
        loc = 1
        for k in range(nloc):
            if np.array_equal(x, xloc[k]):
                loc = 0
                break
        return loc

    def addloc(self, nloc: int, xloc: List[float], x: float) -> Tuple[int, List[float]]:
        """
        Adding a location

        :param nloc:
        :param xloc:
        :param x:
        :return: locations including the added one
        """
        nloc = nloc + 1
        xloc.append(copy.deepcopy(x))
        return nloc, xloc

    def chrelerr(self, fbest: float, stop: List[Union[int, float]]) -> int:
        """
        Performing the chrelerr

        :param fbest:
        :param stop:
        :return: flags
        """
        fglob = stop[1]
        if fbest - fglob <= max(stop[0] * abs(fglob), stop[2]):
            flag = 0
        else:
            flag = 1

        return flag

    def chvtr(self, f: float, vtr: float) -> int:
        """
        Performing te chvtr function

        :param f:
        :param vtr:
        :return: flag
        """
        if f <= vtr:
            flag = 0
        else:
            flag = 1

        return flag

    def fbestloc(self, fmi: List[float], fbest: float, xmin: List[float],
                 xbest: float, nbasket0: int, stop: List[Union[float, int]]) -> Tuple[float, float]:
        """
        The fbestloc function of MCS

        :param fmi:
        :param fbest:
        :param xmin:
        :param xbest:
        :param nbasket0:
        :param stop:
        :return:
        """
        if fmi[nbasket0] < fbest:
            fbest = copy.deepcopy(fmi[nbasket0])
            xbest = copy.deepcopy(xmin[nbasket0])
        return fbest, xbest

    def splrnk(self, n: int, n0: NDArray[np.float64], p: NDArray[np.int32], x: NDArray[np.float64],
               y: NDArray[np.float32]) -> Tuple[int, float]:
        """
        Determines the splitting index and splitting value for splitting a box by rank

        :param n: dimension of the problem
        :param p: ranking of estimated variability of the function in the different coordinates
        :param x: base vertex of the box
        :param y: opposite vertex of the box
        :return : splitting index and value at splitting point
        """

        isplit = 0
        n1 = n0[0]
        p1 = p[0]
        for i in range(1, n):
            if n0[i] < n1 or (n0[i] == n1 and p[i] < p1):
                isplit = i
                n1 = n0[i]
                p1 = p[i]
        if n1 > 0:
            splval = self.split2(x[isplit], y[isplit])
        else:
            splval = np.Inf
        return isplit, splval

    def genbox(self, par: int, level0: int, nchild: int, f0: float) -> Tuple[int, int, int, float]:
        """
        Function that generates a box

        :param par:
        :param level0:
        :param nchild:
        :param f0: inital function value
        :return: Metrics and parameters from generating the box
        """
        ipar = par
        level = level0
        ichild = nchild
        f = f0
        return ipar, level, ichild, f

    def vertex(self, j: int, n: int, u: List[Union[int, float]], v: List[Union[int, float]],
               v1: NDArray[np.float64], x0: NDArray[np.float64], f0: NDArray[np.float64],
               ipar: NDArray[np.int32], isplit: NDArray[np.int32], ichild: NDArray[np.int32],
               z: NDArray[np.float64], f: NDArray[np.float64], l: NDArray[np.int32],
               L: NDArray[np.int32]):
        """
        Vertex function

        :param j: label
        :param n:
        :param u: the initial lower bound ("lower corner" in 3D)
        :param v: the initial upper bound ("upper corner" in 3D)
        :param v1:
        :param x0: initial position
        :param f0: inital function value
        :param ipar:
        :param isplit:
        :param ichild:
        :param z:
        :param f:
        :param l: Indication of the mid point
        :param L: Indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        """
        x = np.multiply(np.Inf, np.ones(n))
        y = np.multiply(np.Inf, np.ones(n))
        x1 = np.multiply(np.Inf, np.ones(n))
        x2 = np.multiply(np.Inf, np.ones(n))
        f1 = np.zeros(n)
        f2 = np.zeros(n)

        n0 = np.zeros(n)
        fold = f[0, j]
        m = j

        while m > 0:
            if isplit[ipar[m]] < 0:
                i = int(abs(isplit[ipar[m]])) - 1
            else:
                i = int(abs(isplit[ipar[m]]))

            n0[i] = n0[i] + 1

            if ichild[m] == 1:
                if x[i] == np.Inf or x[i] == z[0, ipar[m]]:
                    x[i], x1[i], x2[i], f1[i], f2[i] = self.vert1(1, z[:, ipar[m]],
                                                                  f[:, ipar[m]], x1[i], x2[i], f1[i], f2[i])
                else:
                    f1, f2, fold = self.updtf(n, i, x1, x2, f1, f2, fold, f[0, ipar[m]])
                    x1[i], x2[i], f1[i], f2[i] = self.vert2(0, x[i], z[:, ipar[m]],
                                                            f[:, ipar[m]], x1[i], x2[i], f1[i], f2[i])
            elif ichild[m] >= 2:
                f1, f2, fold = self.updtf(n, i, x1, x2, f1, f2, fold, f[0, ipar[m]])
                if x[i] == np.Inf or x[i] == z[1, ipar[m]]:
                    x[i], x1[i], x2[i], f1[i], f2[i] = self.vert1(0, z[:, ipar[m]],
                                                                  f[:, ipar[m]], x1[i], x2[i], f1[i], f2[i])
                else:
                    x1[i], x2[i], f1[i], f2[i] = self.vert2(1, x[i], z[:, ipar[m]],
                                                            f[:, ipar[m]], x1[i], x2[i], f1[i], f2[i])

            if 1 <= ichild[m] and ichild[m] <= 2 and y[i] == np.Inf:
                y[i] = self.split1(z[0, ipar[m]], z[1, ipar[m]], f[0, ipar[m]], f[1, ipar[m]])

            if ichild[m] < 0:
                if u[i] < x0[i, 0]:
                    j1 = math.ceil(abs(ichild[m]) / 2)
                    j2 = math.floor(abs(ichild[m]) / 2)
                    if (abs(ichild[m]) / 2 < j1 and j1 > 0) or j1 == L[i] + 1:
                        j3 = -1
                    else:
                        j3 = 1
                else:
                    j1 = math.floor(abs(ichild[m]) / 2) + 1
                    j2 = math.ceil(abs(ichild[m]) / 2)
                    if abs(ichild[m]) / 2 + 1 > j1 and j1 < L[i] + 1:
                        j3 = 1
                    else:
                        j3 = -1
                j1 -= 1
                j2 -= 1

                if int(isplit[ipar[m]]) < 0:
                    k = copy.deepcopy(i)
                else:
                    k = int(z[0, ipar[m]])

                if j1 != l[i] or (x[i] != np.Inf and x[i] != x0[i, l[i]]):
                    f1, f2, fold = self.updtf(n, i, x1, x2, f1, f2, fold, f0[l[i], k])
                if x[i] == np.Inf or x[i] == x0[i, j1]:
                    x[i] = x0[i, j1]
                    if x1[i] == np.Inf:
                        x1[i], x2[i], f1[i], f2[i] = self.vert3(j1, x0[i, :], f0[:, k], L[i], x1[i], x2[i], f1[i],
                                                                f2[i])
                    elif x2[i] == np.Inf and x1[i] != x0[i, j1 + j3]:
                        x2[i] = x0[i, j1 + j3]
                        f2[i] = f2[i] + f0[j1 + j3, k]
                    elif x2[i] == np.Inf:
                        if j1 != 1 and j1 != L[i]:
                            x2[i] = x0[i, j1 - j3]
                            f2[i] = f2[i] + f0[j1 - j3, k]
                        else:
                            x2[i] = x0[i, j1 + 2 * j3]
                            f2[i] = f2[i] + f0[j1 + 2 * j3, k]
                else:
                    if x1[i] == np.Inf:
                        x1[i] = x0[i, j1]
                        f1[i] = f1[i] + f0[j1, k]
                        if x[i] != x0[i, j1 + j3]:
                            x2[i] = x0[i, j1 + j3]
                            f2[i] = f2[i] + f0[j1 + j3, k]
                    elif x2[i] == np.Inf:
                        if x1[i] != x0[i, j1]:
                            x2[i] = x0[i, j1]
                            f2[i] = f2[i] + f0[j1, k]
                        elif x[i] != x0[i, j1 + j3]:
                            x2[i] = x0[i, j1 + j3]
                            f2[i] = f2[i] + f0[j1 + j3, k]
                        else:
                            if j1 != 1 and j1 != L[i]:
                                x2[i] = x0[i, j1 - j3]
                                f2[i] = f2[i] + f0[j1 - j3, k]
                            else:
                                x2[i] = x0[i, j1 + 2 * j3]
                                f2[i] = f2[i] + f0[j1 + 2 * j3, k]
                if y[i] == np.Inf:
                    if j2 == -1:
                        y[i] = u[i]
                    elif j2 == L[i]:
                        y[i] = v[i]
                    else:
                        y[i] = self.split1(x0[i, j2], x0[i, j2 + 1], f0[j2, k], f0[j2 + 1, k])
            m = ipar[m]
        for i in range(n):
            if x[i] == np.Inf:
                x[i] = x0[i, l[i]]
                x1[i], x2[i], f1[i], f2[i] = self.vert3(l[i], x0[i, :], f0[:, i], L[i], x1[i], x2[i], f1[i], f2[i])
            if y[i] == np.Inf:
                y[i] = v1[i]

        return n0, x, y, x1, x2, f1, f2

    def initbox(self, theta0: NDArray[np.float64], f0: NDArray[np.float32], l: NDArray[np.int32],
                L: NDArray[np.int32], istar: NDArray[Union[np.float32, np.float64]], u: List[Union[int, float]],
                v: List[Union[int, float]], isplit: NDArray[np.int32], level: NDArray[np.int32],
                ipar: NDArray[np.int32], ichild: NDArray[np.int32], f: NDArray[np.float32], nboxes: int, prt: int):
        """
        Generates the boxes in the initializaiton procedure

        :param theta0:
        :param f0: inital function value
        :param l: Indication of the mid point
        :param L: Indication of the end point (or total number of partition of the value x in the i'th dimenstion)
        :param istar:
        :param u:
        :param v:
        :param isplit:
        :param level:
        :param ipar:
        :param ichild:
        :param ichild:
        :param f: function value of the splitinhg float value
        :param nboxes: counter for boxes not in the 'shopping bas
        :param prt: print - unsued in this implementation so far
        """
        n = len(u)

        ipar[0] = -1
        level[0] = 1
        ichild[0] = 1

        f[0, 0] = f0[l[0], 0]

        par = 0

        var = np.zeros(n)
        for i in range(n):
            isplit[par] = - i - 1
            nchild = 0
            if theta0[i, 0] > u[i]:
                nboxes = nboxes + 1
                nchild = nchild + 1
                ipar[nboxes], level[nboxes], ichild[nboxes], f[0, nboxes] = \
                    MCSUtils().genbox(par, level[par] + 1, - nchild, f0[0, i])
            if L[i] == 2:
                v1 = v[i]
            else:
                v1 = theta0[i, 2]
            d = self.polint(theta0[i, 0: 3], f0[0: 3, i])
            xl = self.quadmin(u[i], v1, d, theta0[i, 0: 3])
            fl = self.quadpol(xl, d, theta0[i, 0: 3])
            xu = self.quadmin(u[i], v1, - d, theta0[i, 0: 3])
            fu = self.quadpol(xu, d, theta0[i, 0: 3])

            if istar[i] == 0:
                if xl < theta0[i, 0]:
                    par1 = nboxes
                else:
                    par1 = nboxes + 1

            for j in range(L[i]):
                nboxes = nboxes + 1
                nchild = nchild + 1
                if f0[j, i] <= f0[j + 1, i]:
                    s = 1
                else:
                    s = 2
                ipar[nboxes], level[nboxes], ichild[nboxes], f[0, nboxes] = \
                    MCSUtils().genbox(par, level[par] + s, - nchild, f0[j, i])

                if j >= 1:
                    if istar[i] == j:
                        if xl <= theta0[i, j]:
                            par1 = nboxes - 1
                        else:
                            par1 = nboxes
                    if j <= L[i] - 2:
                        d = self.polint(theta0[i, j: j + 1], f0[j: j + 1, i])
                        if j < L[i] - 2:
                            u1 = theta0[i, j + 1]
                        else:
                            u1 = v[i]
                        xl = self.quadmin(theta0[i, j], u1, d, theta0[i, j: j + 1])
                        fl = min(self.quadpol(xl, d, theta0[i, j: j + 1]), fl)
                        xu = self.quadmin(theta0[i, j], u1, -d, theta0[i, j: j + 1])
                        fu = max(self.quadpol(xu, d, theta0[i, j: j + 1]), fu)

                nboxes = nboxes + 1
                nchild = nchild + 1
                ipar[nboxes], level[nboxes], ichild[nboxes], f[0, nboxes] = \
                    MCSUtils().genbox(par, level[par] + 3 - s, -nchild, f0[j + 1, i])
            if theta0[i, L[i]] < v[i]:
                nboxes = nboxes + 1
                nchild = nchild + 1
                ipar[nboxes], level[nboxes], ichild[nboxes], f[0, nboxes] = \
                    MCSUtils().genbox(par, level[par] + 1, -nchild, f0[L[i], i])

            if istar[i] == L[i]:
                if theta0[i, L[i]] < v[i]:
                    if xl <= theta0[i, L[i]]:
                        par1 = nboxes - 1
                    else:
                        par1 = nboxes
                else:
                    par1 = nboxes
            var[i] = fu - fl

            level[par] = 0
            par = par1
        fbest = f0[istar[n - 1], n - 1]
        p = np.zeros(n).astype(int)
        xbest = np.zeros(n)
        for i in range(n):
            p[i] = np.argmax(var)
            var[p[i]] = -1
            xbest[i] = theta0[i, istar[i]]
        return ipar, level, ichild, f, isplit, p, xbest, fbest, nboxes

    def neighbor(self, x: NDArray[np.float32], delta: List[float],
                 u: List[Union[float, int]], v: List[Union[float, int]]) -> Tuple[List[Any], List[Any]]:
        """
        Computes 'neighbors' x1 and x2 of x needed for making triple search
        and building a local quadratic model such that x(i), x1(i), x2(i) are
        pairwise distinct for i = 1,...,n

        :param x: current position
        :param delta: radius of neighboring region
        :param u:
        :param v:
        """
        i1 = [i for i in range(len(x)) if x[i] == u[i]]
        i2 = [i for i in range(len(x)) if x[i] == v[i]]
        x1 = [max(u[i], x[i] - delta[i]) for i in range(len(x))]
        x2 = [min(x[i] + delta[i], v[i]) for i in range(len(x))]
        for i in i1:
            x1[i] = x[i] + 2 * delta[i]
        for i in i2:
            x2[i] = x[i] - 2 * delta[i]
        return x1, x2

    def polint1(self, x: List[float], f: List[float]) -> Tuple[float, float]:
        """
        Quadratic polynomial interpolation

        :param x: positions
        :param f: function values
        :return: g, G
        """
        f13 = (f[2] - f[0]) / (x[2] - x[0])
        f12 = (f[1] - f[0]) / (x[1] - x[0])
        f23 = (f[2] - f[1]) / (x[2] - x[1])
        g = f13 + f12 - f23
        G = 2 * (f13 - f12) / (x[2] - x[1])
        return g, G

    def hessian(self, i: int, k: int, x: List[Union[float, int]], x0: List[Union[float, int]],
                f: float, f0: float, g: NDArray[np.float64], G: NDArray[np.float64]) -> Any:
        """
        Computes the element G(i,k) of the Hessian of the local quadratic model

        :param i:
        :param k:
        :param x: position
        :param x0: initial position
        :param f: function values
        :param f0: inital function value
        :param g:
        :param G:
        """
        h = f - f0 - g[i] * (x[i] - x0[i]) - g[k] * (x[k] - x0[k]) - 0.5 * G[i, i] * (pow((x[i] - x0[i]), 2)) \
            - 0.5 * G[k, k] * pow((x[k] - x0[k]), 2)
        h = h / (x[i] - x0[i]) / (x[k] - x0[k])
        return h

    def get_theta0(self, iinit: int, u: List[Union[float, int]], v: List[Union[float, int]], n: int) \
            -> NDArray[np.float32]:
        """
        Function for obtaining initial position

        :param iinit:
        :param u:
        :param v:
        :param n:
        :return: the initial position theta0
        """
        if iinit == 0:
            theta0 = np.array([])
            theta0 = np.append(theta0, u, axis=0)
            theta0 = np.vstack([theta0, [(i + j) / 2 for i, j in zip(u, v)]])
            theta0 = np.vstack([theta0, v])
            theta0 = theta0.T

        elif iinit == 1:
            theta0 = np.zeros((n, 3))
            for i in range(n):
                if u[i] >= 0:
                    theta0[i, 0] = u[i]
                    theta0[i, 1], theta0[i, 2] = self.subint(u[i], v[i])
                    theta0[i, 1] = 0.5 * (theta0[i, 0] + theta0[i, 2])
                elif v[i] <= 0:
                    theta0[i, 2] = v[i]
                    theta0[i, 1], theta0[i, 0] = self.subint(v[i], u[i])
                    theta0[i, 1] = 0.5 * (theta0[i, 0] + theta0[i, 2])
                else:
                    theta0[i, 1] = 0
                    _, theta0[i, 0] = self.subint(0, u[i])  # type: ignore[name-defined]
                    _, theta0[i, 2] = self.subint(0, v[i])  # type: ignore[name-defined]
        elif iinit == 2:
            theta0 = np.array([])
            theta0 = np.append(theta0, [(i * 5 + j) / 6 for i, j in zip(u, v)])
            theta0 = np.vstack([theta0, [0.5 * (i + j) for i, j in zip(u, v)]])
            theta0 = np.vstack([theta0, [(i + j * 5) / 6 for i, j in zip(u, v)]])
            theta0 = theta0.T

        if np.any(np.isinf(theta0)):
            sys.exit("Error- MCS main: infinities in ititialization list")
        return theta0



from typing import Union
import torch
from torch.distributions import MultivariateNormal
from fnn_w_gaussian import FNNwithGaussian
from fnn_w_linear import FNNwithLinear


class ActorCriticNet(torch.nn.Module):
    """
    Implements an Actor Critic with parameterizable actor and critic and number of shared layers/dim.

    Sub-classing the torch.nn.Module to be able to use high-level API for creating the custom network
    """

    def __init__(self, input_dim: int, output_dim: int, hidden_dim: int, num_hidden_layers: int = 2,
                 hidden_activation: str = "ReLU", actor: Union[torch.nn.Module, None] = None,
                 critic: Union[None, torch.nn.Module] = None):
        """
        Builds the model

        :param input_dim: the input dimension
        :param output_dim: the output dimension
        :param hidden_dim: the hidden dimension
        :param num_hidden_layers: the number of hidden layers
        :param hidden_activation: hidden activation type
        """
        super(ActorCriticNet, self).__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self.num_hidden_layers = num_hidden_layers
        self.num_layers = num_hidden_layers + 2
        self.hidden_activation = hidden_activation
        self.actor = actor
        self.critic = critic

        # Define layers of FNN
        self.layers = torch.nn.ModuleList()

        # Input layer
        self.layers.append(torch.nn.Linear(input_dim, hidden_dim))
        self.layers.append(self.get_hidden_activation())

        # Hidden Layers
        for i in range(self.num_hidden_layers):
            self.layers.append(torch.nn.Linear(hidden_dim, hidden_dim))
            self.layers.append(self.get_hidden_activation())

        # Shared Output layer
        self.layers.append(torch.nn.Linear(hidden_dim, self.output_dim))

    def get_hidden_activation(self):
        """
        Interprets the hidden activation

        :return: the hidden activation function
        """
        if self.hidden_activation == "ReLU":
            return torch.nn.ReLU()
        elif self.hidden_activation == "LeakyReLU":
            return torch.nn.LeakyReLU()
        elif self.hidden_activation == "LogSigmoid":
            return torch.nn.LogSigmoid()
        elif self.hidden_activation == "PReLU":
            return torch.nn.PReLU()
        elif self.hidden_activation == "Sigmoid":
            return torch.nn.Sigmoid()
        elif self.hidden_activation == "Softplus":
            return torch.nn.Softplus()
        elif self.hidden_activation == "Tanh":
            return torch.nn.Tanh()
        else:
            raise ValueError("Activation type: {} not recognized".format(self.hidden_activation))

    def forward(self, x, actor_only=True):
        """
        Forward propagation

        :param x: input tensor
        :return: Output prediction
        """
        if self.actor is None:
            raise ValueError("Actor cannot be None")
        if self.critic is None:
            raise ValueError("Critic cannot be None")
        y = x
        for i in range(len(self.layers)):
            y = self.layers[i](y)
        latent_shared = y
        action_mean, action_std = self.actor(latent_shared)
        if not actor_only:
            values = self.critic(latent_shared)
            return action_mean, action_std, values
        return action_mean, action_std


def test() -> None:
    """
    A basic test-case to verify that the model can fit some randomly generated data

    :return: None
    """
    # Constants
    input_dim = 8
    shared_output_dim = 400
    shared_hidden_dim = 400
    shared_hidden_layers = 2
    batch_size = 64
    actor_output_dim = 2
    actor_hidden_dim = 200
    actor_hidden_layers = 2
    actor_activation = "ReLU"
    critic_output_dim = 1
    critic_hidden_dim = 200
    critic_hidden_layers = 2
    critic_activation = "ReLU"

    # Create model
    actor = FNNwithGaussian(shared_output_dim, actor_output_dim, actor_hidden_dim,
                            num_hidden_layers=actor_hidden_layers, hidden_activation=actor_activation)
    critic = FNNwithLinear(shared_output_dim, critic_output_dim, critic_hidden_dim,
                           num_hidden_layers=critic_hidden_layers, hidden_activation=critic_activation)
    model = ActorCriticNet(input_dim, shared_output_dim, shared_hidden_dim, num_hidden_layers=shared_hidden_layers,
                           actor=actor, critic=critic)

    # Create random Tensors to hold inputs and outputs
    x = torch.randn(batch_size, input_dim)
    # y = torch.randn(batch_size, shared_output_dim)

    # Construct our loss function and an Optimizer. The call to model.parameters()
    # in the SGD constructor will contain the learnable parameters of the layers in the model
    criterion = torch.nn.MSELoss(reduction='sum')
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-4)
    for t in range(20000):
        # Forward pass: Compute predicted y by passing x to the model
        action_mean, action_std, values = model(x, actor_only=False)

        # Create Gaussian
        cov_mat = torch.diag_embed(action_std)
        dist = MultivariateNormal(action_mean, cov_mat)
        actions = dist.sample()
        print(f"actions: {actions}")
        actor_pred = action_mean

        # Compute and print loss
        y_actor = torch.tensor([0.2, 0.9])
        y_critic = torch.tensor([0.5])
        actor_loss = criterion(actor_pred, y_actor)
        critic_loss = criterion(values, y_critic)
        total_loss = actor_loss + critic_loss
        if t % 100 == 99:
            print(f"step: {t}, actor loss:{actor_loss.item()}, critic loss:{critic_loss.item()}, "
                  f"total loss:{total_loss.tiem()}")

        # Zero gradients, perform a backward pass, and update the weights.
        optimizer.zero_grad()
        total_loss.backward()
        optimizer.step()


if __name__ == '__main__':
    test()



from typing import List, Union
from abc import abstractmethod


class Node:
    """
    Abstract node type, represents a node in the lookahead tree
    """

    def __init__(self, id: int, history: List[int], parent=None, value: float = -2000, visit_count: int = 0,
                 observation: int = -1, action: int = -1) -> None:
        """
        Initializes the node

        :param id: the id of the node
        :param history: the history of the node
        :param parent: the parent node
        :param V: the value of the node
        :param visit_count: the visit count of the node
        :param observation: the observation of the node
        :param action: the action of the node
        """
        self.history = history
        self.value = value
        self.visit_count = visit_count
        self.id = id
        self.parent = parent
        self.children: List["Node"] = []
        self.observation = observation
        self.action = action

    @abstractmethod
    def add_child(self, node: "Node") -> None:
        """
        Method that adds a child to the node. Should be implemented by classes that inherit from this class

        :param node: the node to add
        :return: None
        """
        pass

    @abstractmethod
    def get_child(self, key: int) -> Union["Node", None]:
        """
        Method that gets the child to the node. Should be implemented by classes that inherit from this class

        :param key: the key to identify the child
        :return: the child
        """
        pass



from typing import List, Optional, Tuple, Any
import math
import time
import os
import numpy as np
import numpy.typing as npt
import gymnasium as gym
import pulp
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.alpha_vectors_policy import AlphaVectorsPolicy
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_agents.agents.base.base_agent import BaseAgent
import csle_agents.constants.constants as agents_constants
import csle_agents.common.pruning as pruning


class HSVIAgent(BaseAgent):
    """
    Heuristic-Search Value Iteration for POMDPs (Trey Smith and Reid Simmons, 2004) Agent
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True,
                 env: Optional[BaseEnv] = None):
        """
        Initializes the HSVI agent

        :param simulation_env_config: configuration of the simulation environment
        :param experiment_config: the experiment configuration
        :param training_job: an existing training job to use (optional)
        :param save_to_metastore: boolean flag whether to save the execution to the metastore
        :param env: the gym environment for training
        """
        super().__init__(simulation_env_config=simulation_env_config, emulation_env_config=None,
                         experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.HSVI
        self.training_job = training_job
        self.save_to_metastore = save_to_metastore
        self.env = env

    def train(self) -> ExperimentExecution:
        """
        Runs the value iteration algorithm to compute V*

        :return: the results
        """
        pid = os.getpid()

        # Initialize metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.HSVI.WIDTHS)
        exp_result.plot_metrics.append(agents_constants.HSVI.LB_SIZES)
        exp_result.plot_metrics.append(agents_constants.HSVI.UB_SIZES)
        exp_result.plot_metrics.append(agents_constants.HSVI.INITIAL_BELIEF_VALUES)

        descr = f"Computation of V* with the HSVI algorithm using " \
                f"simulation:{self.simulation_env_config.name}"

        for seed in self.experiment_config.random_seeds:
            exp_result.all_metrics[seed] = {}
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
            exp_result.all_metrics[seed][agents_constants.HSVI.WIDTH] = []
            exp_result.all_metrics[seed][agents_constants.HSVI.UB_SIZE] = []
            exp_result.all_metrics[seed][agents_constants.HSVI.LB_SIZE] = []

        if self.env is None:
            self.env = gym.make(self.simulation_env_config.gym_env_name,
                                config=self.simulation_env_config.simulation_env_input_config)

        # Initialize training job
        if self.training_job is None:
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=None, simulation_traces=[],
                num_cached_traces=0,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
                self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Initialize execution result
        ts = time.time()
        emulation_name = None
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        simulation_name = self.simulation_env_config.name
        self.exp_execution = ExperimentExecution(result=exp_result, config=self.experiment_config, timestamp=ts,
                                                 emulation_name=emulation_name, simulation_name=simulation_name,
                                                 descr=descr, log_file_path=self.training_job.log_file_path)
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
            self.exp_execution.id = exp_execution_id

        for seed in self.experiment_config.random_seeds:
            ExperimentUtil.set_seed(seed)
            exp_result = self.hsvi_algorithm(exp_result=exp_result, seed=seed)

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                if type(value_vectors[0][0]) is int or type(value_vectors[0][0]) is float \
                        or type(value_vectors[0][0]) is np.int64 or type(value_vectors[0][0]) is np.float64:
                    seed_values = []
                    for seed_idx in range(len(self.experiment_config.random_seeds)):
                        seed_values.append(value_vectors[seed_idx][i])
                    avg = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0]
                    if not math.isnan(avg):
                        avg_metrics.append(avg)
                    ci = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1]
                    if not math.isnan(ci):
                        std_metrics.append(ci)
                    else:
                        std_metrics.append(-1)
                else:
                    avg_metrics.append(-1)
                    std_metrics.append(-1)
                exp_result.avg_metrics[metric] = avg_metrics
                exp_result.std_metrics[metric] = std_metrics

                ts = time.time()
        self.exp_execution.timestamp = ts
        self.exp_execution.result = exp_result
        self.training_job.experiment_result = exp_result
        if self.save_to_metastore:
            MetastoreFacade.update_experiment_execution(experiment_execution=self.exp_execution,
                                                        id=self.exp_execution.id)
            MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
        return self.exp_execution

    def hparam_names(self) -> List[str]:
        """
        :return: a list with the hyperparameter names
        """
        return [agents_constants.COMMON.EVAL_BATCH_SIZE, agents_constants.COMMON.CONFIDENCE_INTERVAL,
                agents_constants.COMMON.RUNNING_AVERAGE, agents_constants.COMMON.GAMMA,
                agents_constants.HSVI.TRANSITION_TENSOR,
                agents_constants.HSVI.REWARD_TENSOR, agents_constants.HSVI.OBSERVATION_TENSOR,
                agents_constants.HSVI.OBSERVATION_SPACE, agents_constants.HSVI.STATE_SPACE,
                agents_constants.HSVI.ACTION_SPACE, agents_constants.HSVI.EPSILON,
                agents_constants.HSVI.INITIAL_BELIEF, agents_constants.HSVI.USE_LP,
                agents_constants.HSVI.PRUNE_FREQUENCY, agents_constants.HSVI.SIMULATION_FREQUENCY,
                agents_constants.HSVI.SIMULATE_HORIZON, agents_constants.HSVI.NUMBER_OF_SIMULATIONS]

    def hsvi_algorithm(self, exp_result: ExperimentResult, seed: int) -> ExperimentResult:
        """
        Runs

        :param exp_result: the experiment result object
        :param seed: the random seed
        :return: the updated experiment result
        """
        discount_factor = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        T = self.experiment_config.hparams[agents_constants.HSVI.TRANSITION_TENSOR].value
        R = self.experiment_config.hparams[agents_constants.HSVI.REWARD_TENSOR].value
        Z = self.experiment_config.hparams[agents_constants.HSVI.OBSERVATION_TENSOR].value
        O = self.experiment_config.hparams[agents_constants.HSVI.OBSERVATION_SPACE].value
        S = self.experiment_config.hparams[agents_constants.HSVI.STATE_SPACE].value
        A = self.experiment_config.hparams[agents_constants.HSVI.ACTION_SPACE].value
        epsilon = self.experiment_config.hparams[agents_constants.HSVI.EPSILON].value
        b0 = self.experiment_config.hparams[agents_constants.HSVI.INITIAL_BELIEF].value
        lp = self.experiment_config.hparams[agents_constants.HSVI.USE_LP].value
        prune_frequency = self.experiment_config.hparams[agents_constants.HSVI.PRUNE_FREQUENCY].value
        simulation_frequency = self.experiment_config.hparams[agents_constants.HSVI.SIMULATION_FREQUENCY].value
        simulate_horizon = self.experiment_config.hparams[agents_constants.HSVI.SIMULATE_HORIZON].value
        number_of_simulations = self.experiment_config.hparams[agents_constants.HSVI.NUMBER_OF_SIMULATIONS].value
        Logger.__call__().get_logger().info(f"Starting the HSVI algorithm, epsilon:{epsilon}, "
                                            f"discount_factor: {discount_factor}, lp_nf: {lp}, "
                                            f"prune_frequency: {prune_frequency}, "
                                            f"simulation_frequency: {simulation_frequency}, "
                                            f"number_of_simulations: {number_of_simulations}, "
                                            f"simulate_horizon: {simulate_horizon}, b0: {b0}")
        alpha_vectors, avg_returns, running_avg_returns, widths, ub_sizes, lb_sizes, initial_belief_values = self.hsvi(
            T=np.array(T), R=np.array(R), O=np.array(O), S=np.array(S), Z=np.array(Z), A=np.array(A),
            gamma=discount_factor, b0=b0, epsilon=epsilon, lp=lp, prune_frequency=prune_frequency,
            simulation_frequency=simulation_frequency, simulate_horizon=simulate_horizon,
            number_of_simulations=number_of_simulations)
        exp_result.all_metrics[seed][agents_constants.HSVI.WIDTHS] = widths
        exp_result.all_metrics[seed][agents_constants.HSVI.LB_SIZES] = lb_sizes
        exp_result.all_metrics[seed][agents_constants.HSVI.UB_SIZES] = ub_sizes
        exp_result.all_metrics[seed][agents_constants.HSVI.INITIAL_BELIEF_VALUES] = initial_belief_values
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = avg_returns
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = running_avg_returns
        alpha_vec_policy = AlphaVectorsPolicy(
            player_type=self.experiment_config.player_type,
            actions=self.simulation_env_config.joint_action_space_config.action_spaces[
                self.experiment_config.player_idx].actions,
            states=self.simulation_env_config.state_space_config.states,
            alpha_vectors=alpha_vectors, agent_type=self.experiment_config.agent_type, avg_R=avg_returns[-1],
            simulation_name=self.simulation_env_config.name, transition_tensor=T, reward_tensor=R)
        exp_result.policies[seed] = alpha_vec_policy
        return exp_result

    def hsvi(self, O: npt.NDArray[Any], Z: npt.NDArray[Any], R: npt.NDArray[Any], T: npt.NDArray[Any],
             A: npt.NDArray[Any], S: npt.NDArray[Any], gamma: float, b0: npt.NDArray[Any],
             epsilon: float, lp: bool = False, prune_frequency: int = 10,
             simulation_frequency: int = 10, simulate_horizon: int = 10, number_of_simulations: int = 10) \
            -> Tuple[List[List[float]], List[float], List[float], List[float], List[int], List[int], List[float]]:
        """
        Heuristic Search Value Iteration for POMDPs (Trey Smith and Reid Simmons, 2004)

        :param O: set of observations of the POMDP
        :param Z: observation tensor of the POMDP
        :param R: reward tensor of the POMDP
        :param T: transition tensor of the POMDP
        :param A: action set of the POMDP
        :param S: state set of the POMDP
        :param gamma: discount factor
        :param b0: initial belief point
        :param epsilon: accuracy parameter
        :param lp: whether to use LP to compute upper bound values or SawTooth approximation
        :param prune_frequency: how often to prune the upper and lower bounds
        :param simulation_frequency: how frequently to simulate the POMDP to compure rewards of current policy
        :param simulate_horizon: length of simulations to compute rewards
        :param number_of_simulations: number of simulations to estimate reward
        :return: None
        """
        # Initialize upper and lower bound
        lower_bound = self.initialize_lower_bound(R=R, S=S, A=A, gamma=gamma)
        upper_bound = self.initialize_upper_bound(T=T, A=A, S=S, gamma=gamma, R=R)

        # Initialize state
        w = self.width(lower_bound=lower_bound, upper_bound=upper_bound, b=b0, S=S, lp=lp)
        iteration = 0
        cumulative_r = 0.0
        average_rewards: List[float] = []
        widths = []
        ub_sizes = []
        lb_sizes = []
        average_running_rewards: List[float] = []
        initial_belief_value = []

        # HSVI iterations
        while w > epsilon:

            lower_bound, upper_bound = self.explore(
                b=b0, epsilon=epsilon, t=0, lower_bound=lower_bound, upper_bound=upper_bound,
                gamma=gamma, S=S, O=O, R=R, T=T, A=A, Z=Z, lp=lp)
            w = self.width(lower_bound=lower_bound, upper_bound=upper_bound, b=b0, S=S, lp=lp)

            if iteration % simulation_frequency == 0:
                r = 0.0
                for i in range(number_of_simulations):
                    r += self.simulate(horizon=simulate_horizon, b0=b0, lower_bound=np.array(lower_bound), Z=Z,
                                       R=R, gamma=gamma, T=T, A=A, O=O, S=S)
                cumulative_r = r / number_of_simulations

            if iteration > 1 and iteration % prune_frequency == 0:
                size_before_lower_bound = len(lower_bound)
                size_before_upper_bound = len(upper_bound[0])
                lower_bound = pruning.prune_lower_bound(lower_bound=np.array(lower_bound), S=S)
                upper_bound = self.prune_upper_bound(upper_bound=upper_bound, S=S, lp=lp)
                Logger.__call__().get_logger().debug(
                    f"[HSVI] Pruning, LB before:{size_before_lower_bound}, after:{len(lower_bound)}, "
                    f"UB before: {size_before_upper_bound},after:{len(upper_bound[0])}")

            initial_belief_V_star_upper = self.upper_bound_value(upper_bound=upper_bound, b=b0, S=S, lp=lp)
            initial_belief_V_star_lower = self.lower_bound_value(lower_bound=lower_bound, b=b0, S=S)
            iteration += 1

            if iteration % self.experiment_config.log_every == 0 and iteration > 0:
                Logger.__call__().get_logger().info(f"[HSVI] iteration: {iteration}, width: {w}, epsilon: {epsilon}, "
                                                    f"R: {cumulative_r}, "
                                                    f"UB size:{len(upper_bound[0])}, LB size:{len(lower_bound)},"
                                                    f"Upper V*[b0]: {initial_belief_V_star_upper},"
                                                    f"Lower V*[b0]:{initial_belief_V_star_lower}")

            average_rewards.append(cumulative_r)
            running_avg_J = ExperimentUtil.running_average(
                average_rewards, self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
            average_running_rewards.append(running_avg_J)
            widths.append(w)
            ub_sizes.append(len(upper_bound[0]))
            lb_sizes.append(len(lower_bound))
            initial_belief_value.append(initial_belief_V_star_lower)

        return (list(lower_bound), average_rewards, average_running_rewards, widths, ub_sizes, lb_sizes,
                initial_belief_value)

    def explore(self, b: npt.NDArray[Any], epsilon: float, t: int, lower_bound: List[int],
                upper_bound: Tuple[List[int], List[int]],
                gamma: float, S: npt.NDArray[Any], O: npt.NDArray[Any], Z: npt.NDArray[Any], R: npt.NDArray[Any],
                T: npt.NDArray[Any], A: npt.NDArray[Any], lp: bool) -> Tuple[List[int], Tuple[List[int], List[int]]]:
        """
        Explores the POMDP tree

        :param b: current belief
        :param epsilon: accuracy parameter
        :param t: the current depth of the exploration
        :param lower_bound: the lower bound on the value function
        :param upper_bound: the upper bound on the value function
        :param gamma: discount factor
        :param S: set of states
        :param O: set of observations
        :param Z: observation tensor
        :param R: reward tensor
        :param T: transition tensor
        :param A: set of actions
        :param lp: whether to use LP to compute upper bound values
        :return: new lower and upper bounds
        """
        w = self.width(lower_bound=lower_bound, upper_bound=upper_bound, b=b, S=S, lp=lp)
        print(f"explore, w:{w}, ep:{epsilon * math.pow(gamma, -t)}")
        if (gamma > 0 and w <= epsilon * math.pow(gamma, -t)) or (w <= epsilon):
            return lower_bound, upper_bound

        # Determine a*
        a_Q_vals = []
        for a in A:
            upper_Q = self.q(b=b, a=a, lower_bound=lower_bound, upper_bound=upper_bound, S=S, O=O, Z=Z,
                             R=R, gamma=gamma, T=T, upper=True, lp=lp)
            a_Q_vals.append(upper_Q)
        a_star = int(np.argmax(np.array(a_Q_vals)))

        # Determine o*
        o_vals = []
        for o in O:
            if self.observation_possible(o=o, b=b, Z=Z, T=T, S=S, a=a_star):
                new_belief = self.next_belief(o=o, a=a_star, b=b, S=S, Z=Z, T=T)
                o_val = (self.p_o_given_b_a(o=o, b=b, a=a_star, S=S, Z=Z, T=T) *
                         self.excess(lower_bound=lower_bound, upper_bound=upper_bound, b=new_belief, S=S,
                                     epsilon=epsilon, gamma=gamma, t=(t + 1), lp=lp))
                o_vals.append(o_val)
            else:
                o_vals.append(-np.inf)
        o_star = int(np.argmax(np.array(o_vals)))

        b_prime = self.next_belief(o=o_star, a=a_star, b=b, S=S, Z=Z, T=T)
        lower_bound, upper_bound = self.explore(b=b_prime, epsilon=epsilon, t=t + 1, lower_bound=lower_bound,
                                                upper_bound=upper_bound, gamma=gamma, S=S, O=O, R=R, T=T, A=A,
                                                Z=Z, lp=lp)

        lower_bound, upper_bound = \
            self.local_updates(lower_bound=lower_bound, upper_bound=upper_bound, b=b, A=A, S=S, Z=Z, O=O, R=R,
                               T=T, gamma=gamma, lp=lp)

        return lower_bound, upper_bound

    def initialize_lower_bound(self, R: npt.NDArray[Any], S: npt.NDArray[Any], A: npt.NDArray[Any],
                               gamma: float) -> List[Any]:
        """
        Initializes the lower bound

        :param R: reward tensor
        :param S: set of states
        :param A: set of actions
        :param gamma: discount factor
        :return: the initialized lower bound
        """
        vals_1 = []
        for a in A:
            vals_2 = []
            for s in S:
                vals_2.append(R[a][s] / (1 - gamma))
            vals_1.append(min(vals_2))
        R_underbar = max(vals_1)
        alpha_vector = np.zeros(len(S))
        alpha_vector.fill(R_underbar)
        lower_bound = []
        lower_bound.append(alpha_vector)
        return lower_bound

    def initialize_upper_bound(self, T: npt.NDArray[Any], A: npt.NDArray[Any], S: npt.NDArray[Any], gamma: float,
                               R: npt.NDArray[Any]) -> Tuple[List[Any], List[Any]]:
        """
        Initializes the upper bound

        :param T: the transition tensor
        :param A: the set of actions
        :param S: the set of states
        :param R: the reward tensor
        :param gamma: the discount factor
        :return: the initialized upper bound
        """
        V, pi = self.vi(T=T, num_states=len(S), num_actions=len(A), R=R, theta=0.0001, discount_factor=gamma)
        point_set = []
        for s in S:
            b = self.generate_corner_belief(s=s, S=S)
            point_set.append([b, V[s]])
        return (point_set, point_set.copy())

    def generate_corner_belief(self, s: int, S: npt.NDArray[Any]):
        """
        Generate the corner of the simplex that corresponds to being in some state with probability 1

        :param s: the state
        :param S: the set of States
        :return: the corner belief corresponding to state s
        """
        b = np.zeros(len(S))
        b[s] = 1
        return b

    def local_updates(self, lower_bound: List[Any], upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any],
                      A: npt.NDArray[Any], S: npt.NDArray[Any], O: npt.NDArray[Any], R: npt.NDArray[Any],
                      T: npt.NDArray[Any], gamma: float,
                      Z: npt.NDArray[Any], lp: bool) -> Tuple[List[Any], Tuple[List[Any], List[Any]]]:
        """
        Perform local updates to the upper and  lower bounds for the given belief in the heuristic-search-exploration

        :param lower_bound: the lower bound on V
        :param upper_bound: the upper bound on V
        :param b: the current belief point
        :param A: the set of actions
        :param S: the set of states
        :param O: the set of observations
        :param R: the reward tensor
        :param T: the transition tensor
        :param gamma: the discount factor
        :param Z: the set of observations
        :param lp: a boolean flag whether to use LP to compute upper bound beliefs
        :return: The updated lower and upper bounds
        """
        new_upper_bound = self.local_upper_bound_update(upper_bound=upper_bound, b=b, A=A, S=S, O=O, R=R, T=T,
                                                        gamma=gamma, Z=Z, lp=lp)
        new_lower_bound = self.local_lower_bound_update(lower_bound=lower_bound, b=b, Z=Z, A=A, O=O, S=S, T=T,
                                                        R=R, gamma=gamma)
        return new_lower_bound, new_upper_bound

    def local_upper_bound_update(self, upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any],
                                 A: npt.NDArray[Any], S: npt.NDArray[Any], O: npt.NDArray[Any], R: npt.NDArray[Any],
                                 T: npt.NDArray[Any], gamma: float, Z: npt.NDArray[Any], lp: bool) \
            -> Tuple[List[Any], List[Any]]:
        """
        Performs a local update to the upper bound during the heuristic-search exploration

        :param upper_bound: the upper bound to update
        :param b: the current belief point
        :param A: the set of actions
        :param S: the set of states
        :param O: the set of observations
        :param R: the reward tensor
        :param T: the transition tensor
        :param gamma: the discount factor
        :param Z: the set of observations
        :param lp: whether or not to use LP to compute upper bound beliefs
        :return: the updated upper bound
        """
        b, new_val = self.upper_bound_backup(upper_bound=upper_bound, b=b, A=A, S=S, Z=Z, O=O, R=R, T=T,
                                             gamma=gamma, lp=lp)
        upper_bound_point_set, corner_points = upper_bound
        upper_bound_point_set.append([b, new_val])
        new_corner_points = self.update_corner_points(corner_points=corner_points, new_point=(b, new_val))
        upper_bound = (upper_bound_point_set, new_corner_points)
        return upper_bound

    def update_corner_points(self, corner_points: List[Any], new_point: Tuple[npt.NDArray[Any], float]) -> List[Any]:
        """
        (Maybe) update the corner points of the upper bound

        :param corner_points: the current set of corner points
        :param new_point: the new point to add to the upper bound
        :return: the new set of corner points
        """
        new_corner_points = []
        for cp in corner_points:
            corner_match = True
            for i in range(len(cp[0])):
                if cp[0][i] != new_point[0][i]:
                    corner_match = False
            if corner_match:
                new_corner_points.append((cp[0], new_point[1]))
            else:
                new_corner_points.append(cp)
        return new_corner_points

    def local_lower_bound_update(self, lower_bound: List[Any], b: npt.NDArray[Any], A: npt.NDArray[Any],
                                 O: npt.NDArray[Any], Z: npt.NDArray[Any], S: npt.NDArray[Any], T: npt.NDArray[Any],
                                 R: npt.NDArray[Any], gamma: float) -> List[Any]:
        """
        Performs a local update to the lower bound given a belief point in the heuristic search

        :param lower_bound: the current lower bound
        :param b: the current belief point
        :param A: the set of actions
        :param O: the set of observations
        :param Z: the observation tensor
        :param S: the set of states
        :param T: the transition tensor
        :param R: the reward tensor
        :param gamma: the discount factor
        :return: the updated lower bound
        """
        beta = self.lower_bound_backup(lower_bound=lower_bound, b=b, A=A, Z=Z, O=O, S=S, T=T, R=R, gamma=gamma)
        if not pruning.check_duplicate(np.array(lower_bound), beta):
            lower_bound.append(beta)
        return lower_bound

    def lower_bound_backup(self, lower_bound: List[Any], b: npt.NDArray[Any], A: npt.NDArray[Any], O: npt.NDArray[Any],
                           Z: npt.NDArray[Any], S: npt.NDArray[Any], T: npt.NDArray[Any], R: npt.NDArray[Any],
                           gamma: float) -> npt.NDArray[Any]:
        """
        Generates a new alpha-vector for the lower bound

        :param lower_bound: the current lower bound
        :param b: the current belief point
        :param A: the set of actions
        :param O: the set of observations
        :param Z: the observation tensor
        :param S: the set of states
        :param T: the transition tensor
        :param R: the reward tensor
        :param gamma: the discount factor
        :return: the new alpha vector
        """
        max_beta_a_o_alpha_vecs = []
        for a in A:
            max_beta_a_vecs = []
            for o in O:
                if self.observation_possible(o=o, b=b, Z=Z, T=T, S=S, a=a):
                    new_belief = np.array(self.next_belief(o=o, a=a, b=b, S=S, Z=Z, T=T))
                    max_alpha_vec = lower_bound[0]
                    max_alpha_val = float("-inf")
                    for alpha_vec in lower_bound:
                        alpha_val = new_belief.dot(alpha_vec)
                        if alpha_val > max_alpha_val:
                            max_alpha_val = alpha_val
                            max_alpha_vec = alpha_vec
                    max_beta_a_vecs.append(max_alpha_vec)
            max_beta_a_o_alpha_vecs.append(max_beta_a_vecs)
        beta_a_vecs = []
        for a in A:
            beta_a_vec = []
            for s in S:
                o_idx = 0
                beta_a_s_val = 0.0
                beta_a_s_val += R[a][s]
                expected_future_vals = 0.0
                for o in O:
                    if self.observation_possible(o=o, b=b, Z=Z, T=T, S=S, a=a):
                        for s_prime in S:
                            expected_future_vals += (max_beta_a_o_alpha_vecs[a][o_idx][s_prime] * Z[a][s_prime][o] *
                                                     T[a][s][s_prime])
                        o_idx += 1
                beta_a_s_val += gamma * expected_future_vals
                beta_a_vec.append(beta_a_s_val)
            beta_a_vecs.append(beta_a_vec)

        beta = beta_a_vecs[0]
        max_val = float("-inf")
        for beta_a_vec in beta_a_vecs:
            val = np.array(beta_a_vec).dot(np.array(b))
            if val > max_val:
                max_val = val
                beta = beta_a_vec

        return np.array(beta)

    def upper_bound_backup(self, upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any], A: npt.NDArray[Any],
                           S: npt.NDArray[Any], O: npt.NDArray[Any], Z: npt.NDArray[Any], R: npt.NDArray[Any],
                           T: npt.NDArray[Any], gamma: float, lp: bool) -> Tuple[npt.NDArray[Any], float]:
        """
        Adds a point to the upper bound

        :param upper_bound: the current upper bound
        :param b: the current belief point
        :param A: the set of actions
        :param S: the set of states
        :param O: the set of observations
        :param Z: the observation tensor
        :param R: the reward tensor
        :param T: the transition tensor
        :param gamma: the discount factor
        :param lp: a boolean flag whether to use LP to compute the upper bound belief
        :return: the new point
        """
        q_vals = []
        for a in A:
            v = 0
            for s in S:
                immediate_r = b[s] * R[a][s]
                expected_future_rew = 0
                for o in O:
                    if self.observation_possible(o=o, b=b, Z=Z, T=T, S=S, a=a):
                        new_belief = self.next_belief(o=o, a=a, b=b, S=S, Z=Z, T=T)
                        for s_prime in S:
                            expected_future_rew += \
                                b[s] * T[a][s][s_prime] * Z[a][s_prime][o] * self.upper_bound_value(
                                    upper_bound=upper_bound, b=new_belief, S=S, lp=lp)
                v += immediate_r + gamma * expected_future_rew
            q_vals.append(v)
        new_val = max(q_vals)
        return b, new_val

    def lp_convex_hull_projection_lp(self, upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any],
                                     S: npt.NDArray[Any]) -> float:
        """
        Reference: (Hauskreht 2000)

        Computes the upper bound belief by performing a projection onto the convex hull of the upper bound,
        it is computed by solving an LP

        :param upper_bound: the upper bound
        :param b: the belief point to compute the value for
        :param S: the set of states
        :return: the upper bound value of the belief point
        """
        upper_bound_point_set, corner_points = upper_bound

        problem = pulp.LpProblem("ConvexHullProjection", pulp.LpMinimize)

        # Convex hull coefficients
        lamb = []
        for i in range(len(upper_bound_point_set)):
            lamb_i = pulp.LpVariable("lambda_" + str(i), lowBound=0, upBound=1, cat=pulp.LpContinuous)
            lamb.append(lamb_i)

        # The objective function
        objective = 0
        for i, point in enumerate(upper_bound_point_set):
            objective += lamb[i] * point[1]
        problem += objective, "Convex hull projection"

        # --- The constraints ---

        # Belief probability constraint
        for j in range(len(S)):
            belief_sum = 0
            for i, point in enumerate(upper_bound_point_set):
                belief_sum += lamb[i] * point[0][j]
            problem += belief_sum == b[j], "BeliefVectorConstraint_" + str(j)

        # Convex Hull constraint
        lambda_weights_sum = 0
        for i in range(len(lamb)):
            lambda_weights_sum += lamb[i]
        problem += lambda_weights_sum == 1, "ConvexHullWeightsSumConstraint"

        # Solve
        problem.solve(pulp.PULP_CBC_CMD(msg=0))

        # Obtain solution
        projected_lamb_coefficients = []
        belief_value = 0
        for i in range(len(upper_bound_point_set)):
            projected_lamb_coefficients.append(lamb[i].varValue)
            belief_value += projected_lamb_coefficients[i] * upper_bound_point_set[i][1]

        return belief_value

    def approximate_projection_sawtooth(self, upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any],
                                        S: npt.NDArray[Any]) -> float:
        """
        Reference: (Hauskreht 2000)

        Performs an approximate projection of the belief onto the convex hull of the upepr bound
        to compute the upper bound value of the belief

        :param upper_bound: the upper bound
        :param b: the belief point
        :param S: the set of states
        :return: the value of the belief point
        """
        upper_bound_point_set, corner_points = upper_bound
        alpha_corner = np.array(list(map(lambda x: x[1], corner_points)))
        # min_val = corner_points_belief_value
        interior_belief_values = []
        for point in upper_bound_point_set:
            interior_belief_values.append(self.interior_point_belief_val(interior_point=point, b=b,
                                                                         alpha_corner=alpha_corner, S=S))
        return min(interior_belief_values)

    def interior_point_belief_val(self, interior_point: Tuple[npt.NDArray[Any], float], b: npt.NDArray[Any],
                                  alpha_corner: npt.NDArray[Any], S: npt.NDArray[Any]) -> float:
        """
        Computes the value induced on the belief point b projected onto the convex hull by a given interior belief point

        :param interior_point: the interior point
        :param b: the belief point
        :param alpha_corner: the alpha vector corresponding to the corners of the belief simplex
        :param S: the set of states
        :return: the value of the belief point induced by the interior point
        """
        min_ratios = []
        for s in S:
            if interior_point[0][s] > 0:
                min_ratio = b[s] / interior_point[0][s]
                min_ratios.append(min_ratio)
            else:
                min_ratios.append(float("inf"))
        min_ratio = min(min_ratios)

        if min_ratio > 1:
            min_ratio = 1

        interior_alpha_corner_dot = alpha_corner.dot(interior_point[0])

        return float(interior_alpha_corner_dot + min_ratio * (interior_point[1] - interior_alpha_corner_dot))

    def upper_bound_value(self, upper_bound: Tuple[List[Any], List[Any]], b: npt.NDArray[Any], S: npt.NDArray[Any],
                          lp: bool = False) -> float:
        """
        Computes the upper bound value of a given belief point

        :param upper_bound: the upper bound
        :param b: the belief point
        :param S: the set of states
        :param lp: boolean flag that decides whether to use LP to compute the upper bound value or not
        :return: the upper bound value
        """

        if lp:
            return self.lp_convex_hull_projection_lp(upper_bound=upper_bound, b=b, S=S)
        else:
            return self.approximate_projection_sawtooth(upper_bound=upper_bound, b=b, S=S)

    def lower_bound_value(self, lower_bound: List[Any], b: npt.NDArray[Any], S: npt.NDArray[Any]) -> float:
        """
        Computes the lower bound value of a given belief point

        :param lower_bound: the lower bound
        :param b: the belief point
        :param S: the set of states
        :return: the lower bound value
        """
        alpha_vals = []
        for alpha_vec in lower_bound:
            sum = 0
            for s in S:
                sum += b[s] * alpha_vec[s]
            alpha_vals.append(sum)
        return max(alpha_vals)

    def next_belief(self, o: int, a: int, b: npt.NDArray[Any], S: npt.NDArray[Any], Z: npt.NDArray[Any],
                    T: npt.NDArray[Any]) -> npt.NDArray[Any]:
        """
        Computes the next belief using a Bayesian filter

        :param o: the latest observation
        :param a: the latest action
        :param b: the current belief
        :param S: the set of states
        :param Z: the observation tensor
        :param T: the transition tensor
        :return: the new belief
        """
        b_prime = np.zeros(len(S))
        for s_prime in S:
            b_prime[s_prime] = self.bayes_filter(s_prime=s_prime, o=o, a=a, b=b, S=S, Z=Z, T=T)
        assert round(sum(b_prime), 5) == 1
        return b_prime

    def bayes_filter(self, s_prime: int, o: int, a: int, b: npt.NDArray[Any], S: npt.NDArray[Any], Z: npt.NDArray[Any],
                     T: npt.NDArray[Any]) -> float:
        """
        A Bayesian filter to compute the belief of being in s_prime when observing o after taking action a in belief b

        :param s_prime: the state to compute the belief of
        :param o: the observation
        :param a: the action
        :param b: the current belief point
        :param S: the set of states
        :param Z: the observation tensor
        :param T: the transition tensor
        :return: b_prime(s_prime)
        """
        norm = 0
        for s in S:
            for s_prime_1 in S:
                prob_1 = Z[a][s_prime_1][o]
                norm += b[s] * prob_1 * T[a][s][s_prime_1]
        if norm == 0:
            return 0
        temp = 0
        for s in S:
            temp += b[s] * T[a][s][s_prime] * Z[a][s_prime][o]
        b_prime = (temp) / norm
        assert b_prime <= 1
        return b_prime

    def p_o_given_b_a(self, o: int, b: npt.NDArray[Any], a: int, S: npt.NDArray[Any], Z: npt.NDArray[Any],
                      T: npt.NDArray[Any]) -> float:
        """
        Computes P[o|a,b]

        :param o: the observation
        :param b: the belief point
        :param a: the action
        :param S: the set of states
        :param Z: the observation tensor
        :param T: the transition tensor
        :return: the probability of observing o when taking action a in belief point b
        """
        prob = 0
        for s in S:
            for s_prime in S:
                prob += b[s] * T[a][s][s_prime] * Z[a][s_prime][o]
        prob = round(prob, 3)
        assert round(prob, 3) <= 1
        return prob

    def excess(self, lower_bound: List[Any], upper_bound: Tuple[List[Any], List[Any]],
               b: npt.NDArray[Any], S: npt.NDArray[Any],
               epsilon: float, gamma: float, t: int, lp: bool) -> float:
        """
        Computes the excess uncertainty  (Trey Smith and Reid Simmons, 2004)

        :param lower_bound: the lower bound
        :param upper_bound: the upper bound
        :param b: the current belief point
        :param S: the set of states
        :param epsilon: the epsilon accuracy parameter
        :param gamma: the discount factor
        :param t: the current exploration depth
        :param lp: whether to use LP or not to compute upper bound belief values
        :return: the excess uncertainty
        """
        w = self.width(lower_bound=lower_bound, upper_bound=upper_bound, b=b, S=S, lp=lp)
        if gamma == 0:
            return w
        else:
            return w - epsilon * math.pow(gamma, -(t))

    def width(self, lower_bound: List[Any], upper_bound: Tuple[List[Any], List[Any]],
              b: npt.NDArray[Any], S: npt.NDArray[Any],
              lp: bool) -> float:
        """
        Computes the bounds width (Trey Smith and Reid Simmons, 2004)

        :param lower_bound: the current lower bound
        :param upper_bound: the current upper bound
        :param b: the current belief point
        :param S: the set of states
        :param lp: boolean flag that decides whether to use LP to compute upper bound belief values
        :return: the width of the bounds
        """
        ub = self.upper_bound_value(upper_bound=upper_bound, b=b, S=S, lp=lp)
        lb = self.lower_bound_value(lower_bound=lower_bound, b=b, S=S)
        return ub - lb

    def q_hat_interval(self, b: npt.NDArray[Any], a: int, S: npt.NDArray[Any], O: npt.NDArray[Any],
                       Z: npt.NDArray[Any], R: npt.NDArray[Any], T: npt.NDArray[Any], gamma: float,
                       lower_bound: List[int], upper_bound: Tuple[List[Any], List[Any]], lp: bool) -> List[float]:
        """
        Computes the interval (Trey Smith and Reid Simmons, 2004)

        :param b: the current belief point
        :param a: the action
        :param S: the set of states
        :param O: the set of observations
        :param Z: the observation tensor
        :param R: the reward tensor
        :param T: the transition tensor
        :param gamma: the discount factor
        :param lower_bound: the lower bound
        :param upper_bound: the upper bound
        :param lp: boolean flag that decides whether to use LP to compute upper bound belief values
        :return: the interval
        """
        upper_Q = self.q(b=b, a=a, lower_bound=lower_bound, upper_bound=upper_bound, S=S, O=O, Z=Z,
                         R=R, gamma=gamma, T=T, upper=True, lp=lp)
        lower_Q = self.q(b=b, a=a, lower_bound=lower_bound, upper_bound=upper_bound, S=S, O=O, Z=Z,
                         R=R, gamma=gamma, T=T, upper=False, lp=lp)
        return [lower_Q, upper_Q]

    def q(self, b: npt.NDArray[Any], a: int, lower_bound: List[int], upper_bound: Tuple[List[int], List[int]],
          S: npt.NDArray[Any], O: npt.NDArray[Any], Z: npt.NDArray[Any], R: npt.NDArray[Any], gamma: float,
          T: npt.NDArray[Any], upper: bool = True, lp: bool = False) -> float:
        """
        Applies the Bellman equation to compute Q values

        :param b: the belief point
        :param a: the action
        :param lower_bound: the lower bound
        :param upper_bound: the upper bound
        :param S: the set of states
        :param O: the set of observations
        :param Z: the observation tensor
        :param R: the reward tensor
        :param gamma: the discount factor
        :param T: the transition tensor
        :param upper: boolean flag that decides whether to use the upper bound or lower bound on V to
                      compute the Q-value
        :param lp: boolean flag that decides whether to use LP to compute upper bound belief values
        :return: the Q-value
        """
        Q_val = 0
        for s in S:
            immediate_r = R[a][s] * b[s]
            expected_future_rew = 0
            for o in O:
                if self.observation_possible(o=o, b=b, Z=Z, T=T, S=S, a=a):
                    new_belief = self.next_belief(o=o, a=a, b=b, S=S, Z=Z, T=T)
                    for s_prime in S:
                        if upper:
                            future_val = self.upper_bound_value(upper_bound=upper_bound, b=new_belief, S=S, lp=lp)
                        else:
                            future_val = self.lower_bound_value(lower_bound=lower_bound, b=new_belief, S=S)
                        expected_future_rew += \
                            b[s] * T[a][s][s_prime] * Z[a][s_prime][o] * future_val
            Q_val += immediate_r + expected_future_rew * gamma
        return Q_val

    def prune_upper_bound(self, upper_bound: Tuple[List[Any], List[Any]], S: npt.NDArray[Any], lp: bool) \
            -> Tuple[List[Any], List[Any]]:
        """
        Prunes the points in the upper bound

        :param upper_bound: the current upper bound
        :param S: the set of states
        :param lp: boolean flag that decides whether to use LP to compute upper bound belief values
        :return: the pruned upper bound
        """

        upper_bound_point_set, corner_points = upper_bound
        pruned_upper_bound_point_set = []

        for point in upper_bound_point_set:
            true_val = self.upper_bound_value(upper_bound=upper_bound, S=S, b=point[0], lp=lp)
            if not (point[1] > true_val):
                pruned_upper_bound_point_set.append(point)

        return pruned_upper_bound_point_set, corner_points

    def simulate(self, horizon: int, b0: npt.NDArray[Any], lower_bound: npt.NDArray[Any],
                 Z: npt.NDArray[Any], R: npt.NDArray[Any], gamma: float,
                 T: npt.NDArray[Any], A: npt.NDArray[Any], O: npt.NDArray[Any], S: npt.NDArray[Any]) -> float:
        """
        Simulates the POMDP to estimate the reward of the greedy policy with respect to the value function represented
        by the lower bound

        :param horizon: the horizon for the simulation
        :param b0: the initial belief
        :param lower_bound: the lower bound which represents the value function
        :param Z: the observation tensor
        :param R: the reward tensor
        :param gamma: the discount factor
        :param T: the transition operator
        :param A: the action set
        :param O: the observation set
        :param S: the set of states
        :return: the cumulative discounted reward
        """
        t = 0
        b = b0
        cumulative_r = 0.0
        while t < horizon:
            q_values = list(map(lambda a: self.q(
                b=b, a=a, lower_bound=list(lower_bound.tolist()), upper_bound=([], []), S=S, O=O, Z=Z,
                R=R, gamma=gamma, T=T, upper=False, lp=False), A))
            a = int(np.argmax(np.array(q_values)))
            r = 0
            for s in S:
                r += R[a][s] * b[s]
            cumulative_r += math.pow(gamma, t) * r
            observation_probabilities = []
            for o in O:
                p = 0
                for s in S:
                    for s_prime in S:
                        p += b[s] * T[a][s][s_prime] * Z[a][s_prime][o]
                observation_probabilities.append(p)
            o = np.random.choice(np.arange(0, len(O)), p=observation_probabilities)
            b = self.next_belief(o=o, a=a, b=b, S=S, Z=Z, T=T)
            t += 1

        return cumulative_r

    def one_step_lookahead(self, state, V, num_actions, num_states, T, discount_factor, R) -> npt.NDArray[Any]:
        """
        Performs a one-step lookahead for value iteration
        :param state: the current state
        :param V: the current value function
        :param num_actions: the number of actions
        :param num_states: the number of states
        :param T: the transition kernel
        :param discount_factor: the discount factor
        :param R: the table with rewards
        :param next_state_lookahead: the next state lookahead table
        :return: an array with lookahead values
        """
        A = np.zeros(num_actions)
        for a in range(num_actions):
            reward = R[a][state]
            for next_state in range(num_states):
                prob = T[a][state][next_state]
                A[a] += prob * (reward + discount_factor * V[next_state])
        return A

    def vi(self, T: npt.NDArray[Any], num_states: int, num_actions: int, R: npt.NDArray[Any],
           theta=0.0001, discount_factor=1.0) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        """
        An implementation of the Value Iteration algorithm
        :param T: the transition kernel T
        :param num_states: the number of states
        :param num_actions: the number of actions
        :param state_to_id: the state-to-id lookup table
        :param HP: the table with hack probabilities
        :param R: the table with rewards
        :param next_state_lookahead: the next-state-lookahead table
        :param theta: convergence threshold
        :param discount_factor: the discount factor
        :return: (greedy policy, value function)
        """
        V = np.zeros(num_states)

        while True:
            # Stopping condition
            delta = 0
            # Update each state...
            for s in range(num_states):
                # Do a one-step lookahead to find the best action
                A = self.one_step_lookahead(s, V, num_actions, num_states, T, discount_factor, R)
                best_action_value = np.max(A)
                # Calculate delta across all states seen so far
                delta = max(delta, np.abs(best_action_value - V[s]))
                # Update the value function. Ref: Sutton book eq. 4.10.
                V[s] = best_action_value

            # Check if we can stop
            if delta < theta:
                break

        # Create a deterministic policy using the optimal value function
        policy = np.zeros([num_states, num_actions * 2])
        for s in range(num_states):
            # One step lookahead to find the best action for this state
            A = self.one_step_lookahead(s, V, num_actions, num_states, T, discount_factor, R)
            best_action = np.argmax(A)
            # Always take the best action
            policy[s, best_action] = 1.0

        return V, policy

    def observation_possible(self, o, b, Z, T, S, a) -> bool:
        """
        Checks if a given observation can be observed when taking a given action in a given state

        :param o: the observation to check
        :param b: the belief to check
        :param Z: the observation tensor
        :param T: the transition tensor
        :param S: the state space
        :param a: the aciton tocheck
        :return: true if possible otherwise fasle
        """
        prob = 0
        for s in S:
            for s_prime in S:
                prob += b[s] * T[a][s][s_prime] * Z[a][s_prime][o]
        return prob > 0
